"""Load, manipulate and save R&S waveform files."""

from __future__ import annotations

import io
import typing
from importlib.metadata import PackageNotFoundError, version

import numpy as np

from . import iqtar, iqw, wv
from .load_interface import LoadInterface
from .meta import Meta
from .parent_storage import ParentStorage
from .save_interface import SaveInterface
from .utility.dsp import (
    calculate_par,
    calculate_peak,
    calculate_rms,
    convert_to_db,
    normalize,
)
from .utility.integer_indexedproperty import IntegerIndexedProperty

if typing.TYPE_CHECKING:
    from pathlib import Path

__version__ = "unknown"
try:
    __version__ = version(__package__)
except PackageNotFoundError:
    # package is not installed
    pass


class RsWaveform(object):
    """RsWaveform class for loading and saving IQ data."""

    def __init__(
        self,
        load: typing.Type[LoadInterface] = wv.Load,
        save: typing.Type[SaveInterface] = wv.Save,
        file: typing.Union[None, str, typing.IO, Path] = None,
        nr_samples: typing.Union[None, int] = None,
        samples_offset: typing.Union[None, int] = None,
        only_meta_data: bool = False,
    ) -> None:
        """Initialize an R&S Waveform class."""
        if not issubclass(load, LoadInterface):
            raise ValueError(f"Loader is not of type {type(LoadInterface)}")
        if not issubclass(save, SaveInterface):
            raise ValueError(f"Saver is not of type {type(SaveInterface)}")
        self.loader = load()
        self.saver = save()
        self._parent_storage = ParentStorage()
        if file:
            if not only_meta_data:
                if nr_samples is None and samples_offset is None:
                    self.load(file)
                elif nr_samples is not None and samples_offset is not None:
                    self.load_in_chunks(file, nr_samples, samples_offset)
                else:
                    raise ValueError(
                        "You need to define both nr_samples and "
                        "samples_offset to load chunks!"
                    )
            else:
                self.load_meta(file)

    def load(self, file: typing.Union[str, typing.IO, Path]) -> None:
        """Load a waveform file with a set loader class."""
        self._parent_storage = self.loader.load(file)

    def load_in_chunks(
        self,
        file: typing.Union[str, typing.IO, Path],
        nr_samples: int,
        samples_offset: int,
    ) -> None:
        """Load a chunk of a waveform file with a set loader class."""
        self._parent_storage = self.loader.load_in_chunks(
            file, nr_samples, samples_offset
        )

    def load_meta(self, file: typing.Union[str, typing.IO, Path]) -> None:
        """Load meta of a waveform file with a set loader class."""
        self._parent_storage = self.loader.load_meta(file)

    def save(
        self,
        file: typing.Union[str, typing.IO, Path],
        scale: float = np.power(2, np.iinfo(np.int16).bits - 1),
    ) -> None:
        """Saver a waveform file with a set saver class."""
        self.saver.save(file, self._parent_storage, scale)

    def tobytes(self) -> bytes:
        """Get waveform file content as bytes."""
        with io.BytesIO() as bio:
            self.save(bio)
            data = bio.getvalue()
        return data

    def frombytes(self, binary_data: typing.Union[bytes, io.BytesIO]) -> None:
        """Load waveform from bytes file content."""
        if isinstance(binary_data, bytes):
            with io.BytesIO(binary_data) as bio:
                self.load(bio)
        else:
            self.load(binary_data)

    @property
    def filename(self) -> str:
        """Get the waveform filename."""
        return self._parent_storage.filename

    @filename.setter
    def filename(self, file_name: str) -> None:
        """Set the waveform filename."""
        self._parent_storage.filename = file_name

    @IntegerIndexedProperty
    def meta(self, index: int = 0) -> Meta:
        """Get the waveform storage metadata."""
        return self._parent_storage.storages[index].meta

    @meta.setter
    def meta_setter(self, index: int, meta: Meta) -> None:
        """Set the waveform storage metadata."""
        self._parent_storage.storages[index].meta = meta

    @IntegerIndexedProperty
    def data(self, index: int = 0) -> np.ndarray:
        """Get the waveform IQ data."""
        return self._parent_storage.storages[index].data

    @data.setter
    def data_setter(self, index: int, data: np.ndarray) -> None:
        """Set the waveform storage IQ data."""
        self._parent_storage.storages[index].data = data

    @property
    def parent_storage(self) -> ParentStorage:
        """Get parent storage."""
        return self._parent_storage


class Iqw(RsWaveform, object):
    """Iqw class for loading and saving IQ data."""

    def __init__(
        self,
        load: typing.Type[LoadInterface] = iqw.Load,
        save: typing.Type[SaveInterface] = iqw.Save,
        file: typing.Union[None, str, typing.IO, Path] = None,
        nr_samples: typing.Union[None, int] = None,
        samples_offset: typing.Union[None, int] = None,
    ) -> None:
        """Initialize a Iqw class."""
        super().__init__(load, save, file, nr_samples, samples_offset, False)

    def save(
        self, file: typing.Union[str, typing.IO, Path], scale: float = 1.0
    ) -> None:
        """Saver an iqw file with a set saver class."""
        self.saver.save(file, self._parent_storage, scale)


class IqTar(RsWaveform, object):
    """Iqtar class for loading and saving IQ data."""

    def __init__(
        self,
        load: typing.Type[LoadInterface] = iqtar.Load,
        save: typing.Type[SaveInterface] = iqtar.Save,
        file: typing.Union[None, str, Path] = None,
        nr_samples: typing.Union[None, int] = None,
        samples_offset: typing.Union[None, int] = None,
        only_meta_data: bool = False,
    ) -> None:
        """Initialize a IQtar class."""
        super().__init__(load, save, file, nr_samples, samples_offset, only_meta_data)

    def save(
        self, file: typing.Union[str, typing.IO, Path], scale: float = 1.0
    ) -> None:
        """Saver an iqw file with a set saver class."""
        self.saver.save(file, self._parent_storage, scale)


__all__ = [
    "RsWaveform",
    "Iqw",
    "IqTar",
    "load_interface",
    "save_interface",
    "calculate_par",
    "calculate_peak",
    "calculate_rms",
    "normalize",
    "convert_to_db",
]



"""Save R&S waveform files."""

from __future__ import annotations

import abc
import typing

from .parent_storage import ParentStorage

if typing.TYPE_CHECKING:
    from pathlib import Path


class SaveInterface(abc.ABC):
    """Load R&S waveform implementation."""

    @abc.abstractmethod
    def save(
        self,
        file: typing.Union[str, typing.IO, Path],
        datas: ParentStorage,
        scale: float = 1.0,
    ) -> None:
        """Abstract save implementation."""
        pass



"""Configuration file for Sphinx documentation builder."""

project = "RsWaveform"
copyright = "Rohde & Schwarz GmbH & Co. KG 2024-present"
author = "Rohde & Schwarz GmbH & Co. KG"
release = version = "0.1.0"

extensions = [
    "sphinx.ext.autodoc",  # Automatically documents modules etc.
    "sphinx.ext.autodoc.typehints",  # Includes typehints for autodoc
    "sphinx.ext.napoleon",  # Parses the docstrings Google style
    "sphinx.ext.viewcode",  # Includes the source code and links in documentation
    "myst_parser",  # Allows the use of .md files instead of .rst files
]
autoclass_content = "both"  # Display class and __init__ docstring
autodoc_typehints = "description"  # Move type annotations to description
myst_enable_extensions = [
    "tasklist",
]
myst_heading_anchors = 3
suppress_warnings = ["myst.header"]
html_theme = "sphinx_rtd_theme"



import io
import os
import tarfile
import tempfile
import typing
from contextlib import contextmanager
from pathlib import Path

import pytest

from RsWaveform.utility.file_handling import (
    read_file_handle,
    read_file_handle_tar,
    write_file_handle,
    write_file_handle_tar,
)


@pytest.fixture
def content() -> bytes:
    return b"test"


@pytest.fixture
def file_name() -> str:
    return "test.txt"


@pytest.fixture
def file(file_name, content) -> str:  # type: ignore[misc]
    with open(file_name, "wb") as fp:
        fp.write(content)
    yield file_name
    if os.path.exists(file_name):
        os.remove(file_name)


@pytest.fixture
def empty_file(file_name: str) -> str:  # type: ignore[misc]
    yield file_name
    if os.path.exists(file_name):
        os.remove(file_name)


@contextmanager
def create_temp_tarfile(
    contents: typing.Dict[str, bytes]
) -> typing.Generator[Path, None, None]:
    """Create a temporary tar file with the given contents."""
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        with tarfile.open(temp_file.name, "w:") as tar:
            for name, data in contents.items():
                info = tarfile.TarInfo(name)
                info.size = len(data)
                tar.addfile(info, io.BytesIO(data))
    yield Path(temp_file.name)
    Path(temp_file.name).unlink()


def test_read_file_with_file_name(file, content):
    read_file(file, content)


def test_read_file_with_file_path(file, content):
    read_file(Path(file), content)


def test_read_file_with_file_io(file, content):
    with open(file, "rb") as ofp:
        read_file(ofp, content)


def test_write_file_with_file_io(empty_file, content):
    with open(empty_file, "wb") as ofp:
        write_file(ofp, content)
    read_file(empty_file, content)


def test_write_file_with_file_name(empty_file, content):
    write_file(empty_file, content)
    read_file(empty_file, content)


def test_write_file_with_file_path(empty_file, content):
    write_file(Path(empty_file), content)
    read_file(Path(empty_file), content)


def write_file(file_name, content):
    with write_file_handle(file_name) as fp:
        fp.write(content)


def read_file(file, content):
    with read_file_handle(file) as fp:
        read = fp.read()
        assert read == content


def test_write_file_with_path(empty_file, content):
    write_file(Path(empty_file), content)
    read_file(Path(empty_file), content)


def test_read_file_handle_tar_existing_tarfile():
    with create_temp_tarfile({"file1.txt": b"Hello, World!"}) as temp_tar:
        with read_file_handle_tar(temp_tar) as tar:
            assert isinstance(tar, tarfile.TarFile)
            assert "file1.txt" in tar.getnames()


def test_read_file_handle_tar_nonexistent_tarfile():
    with pytest.raises(FileNotFoundError):
        with read_file_handle_tar("nonexistent.tar"):
            pass


def test_write_file_handle_tar():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_tar = Path(temp_dir) / "test.tar"
        name = "file1.txt"
        data = b"Hello, World!"
        with write_file_handle_tar(temp_tar) as tar:
            assert isinstance(tar, tarfile.TarFile)
            info = tarfile.TarInfo(name)
            info.size = len(data)
            tar.addfile(info, io.BytesIO(data))

            with read_file_handle_tar(tar) as sub_tar:
                assert "file1.txt" in sub_tar.getnames()
                member = sub_tar.getmember("file1.txt")
                assert member.size == len(data)  # length of "Hello, World!"

        with open(temp_tar, mode="rb") as read_tar:
            with read_file_handle_tar(io.BytesIO(read_tar.read())) as sub_tar:
                assert "file1.txt" in sub_tar.getnames()
                member = sub_tar.getmember("file1.txt")
                assert member.size == len(data)  # length of "Hello, World!"

        with read_file_handle_tar(temp_tar) as tar:
            assert "file1.txt" in tar.getnames()
            member = tar.getmember("file1.txt")
            assert member.size == len(data)  # length of "Hello, World!"



try:
    import msgpack as pickle
    import msgpack_numpy as m

    m.patch()
except ImportError:
    import pickle  # type: ignore

from datetime import datetime

import numpy as np
import pytest

from RsWaveform.meta import Meta
from RsWaveform.parent_storage import ParentStorage
from RsWaveform.storage import Storage, encode_datetime, msgpack_import


@pytest.fixture
def reference_now() -> datetime:
    return datetime.now()


@pytest.fixture
def reference_meta() -> dict:
    return {}


@pytest.fixture
def reference_name() -> str:
    return "test.wv"


@pytest.fixture
def reference_data() -> np.ndarray:
    return np.array([1 + 1j, 1 - 1j, 1 + 0j, -1 + 1j])


@pytest.fixture
def reference_serde(
    reference_now, reference_meta, reference_data, reference_name
) -> dict:
    return {
        "data": reference_data,
        "meta": reference_meta,
        "filename": reference_name,
        "timestamp": reference_now,
    }


def test_creation():
    storage = Storage()
    assert isinstance(storage.data, np.ndarray)
    assert np.array_equal(storage.data, np.zeros((1024,), dtype=np.complex128))
    assert isinstance(storage.meta, Meta)


def test_creation_parent_storage():
    parent_storage = ParentStorage()
    assert isinstance(parent_storage.timestamp, datetime)
    assert parent_storage.timestamp.date() == datetime.now().date()


def test_create_from_serialize(
    reference_serde, reference_meta, reference_now, reference_data, reference_name
):
    if msgpack_import:
        bins = pickle.dumps(reference_serde, default=encode_datetime)
    else:
        bins = pickle.dumps(reference_serde)
    storage = Storage(bins)
    assert np.array_equal(storage.data, reference_data)
    assert storage.meta == reference_meta


def test_serialize(
    reference_serde, reference_meta, reference_now, reference_data, reference_name
):
    storage = Storage()
    storage.data = reference_data
    storage.meta = reference_meta
    storage.filename = reference_name  # type: ignore[attr-defined]
    storage.timestamp = reference_now  # type: ignore[attr-defined]
    bins = storage.serialize()
    if msgpack_import:
        reference = pickle.dumps(reference_serde, default=encode_datetime)
    else:
        reference = pickle.dumps(reference_serde)
    assert bins == reference


def test_deserialize(
    reference_serde, reference_meta, reference_now, reference_data, reference_name
):
    if msgpack_import:
        bins = pickle.dumps(reference_serde, default=encode_datetime)
    else:
        bins = pickle.dumps(reference_serde)
    storage = Storage()
    storage.deserialize(bins)
    assert np.array_equal(storage.data, reference_data)
    assert storage.meta == reference_meta
    assert storage.filename == reference_name  # type: ignore[attr-defined]
    assert storage.timestamp == reference_now  # type: ignore[attr-defined]



import os

import pytest


@pytest.fixture(params=["dummy.wv", "dummy_2.wv"])
def reference_waveform_file_name(request) -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", request.param)


@pytest.fixture
def reference_waveform_file_name_huge() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "huge_dummy.wv")


@pytest.fixture
def reference_waveform_mwv_file_name() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy_mwv.wv")


@pytest.fixture
def reference_iqw_file_name() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy.iqw")


@pytest.fixture
def reference_iqw_file_name_huge() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy_huge.iqw")


@pytest.fixture
def reference_iqtar_file_name() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy.iq.tar")


@pytest.fixture
def reference_iqtar_file_name_huge() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy_huge.iq.tar")


@pytest.fixture
def reference_iqtar_file_name_multi_channel() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy_multi_channel.iq.tar")


@pytest.fixture
def reference_npz_file_name() -> str:
    dirname = os.path.dirname(__file__)
    return os.path.join(os.path.realpath(dirname), "data", "dummy.npz")



import inspect

import pytest

from RsWaveform.meta import Meta
from RsWaveform.meta.defaults import META_IQTAR_DEFAULTS, META_WV_DEFAULTS
from RsWaveform.meta.meta_iqtar import MetaIqtar
from RsWaveform.meta.meta_wv import MetaWv


def test_dictionary_behavior_of_meta_class():
    meta = Meta()
    meta.clear()
    assert len(meta) == 0
    with pytest.raises(KeyError):
        meta["not valid"]
    assert meta.get("not valid") is None
    meta["new key"] = True
    assert meta["new key"]
    assert meta.pop("new key")
    meta.update(test=True)
    assert meta["test"]
    assert meta.items() == meta._items.items()
    assert meta.keys() == meta._items.keys()
    assert list(meta.values()) == list(meta._items.values())
    assert meta == meta.copy()
    assert meta == meta._items
    assert not meta == 1
    assert meta != Meta()
    p = meta.popitem()
    assert p == ("test", True)
    meta.update(test=True)
    del meta["test"]
    assert "test" not in meta
    assert meta.setdefault("test") is None
    assert meta["test"] is None


@pytest.mark.parametrize(
    "meta_wrapper,defaults",
    [(MetaWv, META_WV_DEFAULTS), (MetaIqtar, META_IQTAR_DEFAULTS)],
)
def test_fill_with_defaults(meta_wrapper, defaults):
    meta = meta_wrapper()
    properties = inspect.getmembers(type(meta), lambda o: isinstance(o, property))
    for prop, _ in properties:
        reference = defaults[prop]
        assert getattr(meta, prop) == reference
        assert meta[prop] == reference


@pytest.mark.parametrize(
    "meta_wrapper,defaults,kwargs",
    [
        (
            MetaWv,
            META_WV_DEFAULTS,
            {
                "clock": 2e9,
                "comment": "Test WV",
                "copyright": "my copyright",
                "marker": {"marker list 1": [1, 2]},
            },
        ),
        (
            MetaIqtar,
            META_IQTAR_DEFAULTS,
            {
                "clock": 500e6,
                "comment": "Test IQTAR",
                "center_frequency": 2e9,
                "scalingfactor": 2,
            },
        ),
    ],
)
def test_fill_partly_from_kwargs(meta_wrapper, defaults, kwargs):
    new_defaults = defaults.copy()
    new_defaults.update(kwargs)
    meta = meta_wrapper(**kwargs)
    properties = inspect.getmembers(type(meta), lambda o: isinstance(o, property))
    for prop, _ in properties:
        reference = new_defaults[prop]
        assert getattr(meta, prop) == reference
        assert meta[prop] == reference


def test_meta():
    defaults = META_WV_DEFAULTS.copy()
    defaults.update(META_IQTAR_DEFAULTS)
    kwargs = {
        "clock": 2e9,
        "comment": "Test this",
        "copyright": "my copyright",
        "marker": {"marker list 1": [1, 2]},
        "center_frequency": 2e9,
        "scalingfactor": 2,
    }
    defaults.update(kwargs)
    meta = Meta(**kwargs)  # type: ignore[arg-type]
    properties = inspect.getmembers(type(meta), lambda o: isinstance(o, property))
    for prop, _ in properties:
        reference = defaults[prop]
        assert getattr(meta, prop) == reference
        assert meta[prop] == reference


def test_single_properties():
    meta = Meta(no_defaults=True)
    assert meta.marker == {}
    assert "marker" in meta._items
    assert meta.control_list is None



import numpy as np
import pytest

from RsWaveform.wv.utility.array_to_bytes import (
    pack_bool_array_to_bytes,
    unpack_bytes_to_bool_array,
)


@pytest.fixture
def example_bool_array():
    return np.array([[0, 1], [1, 0], [1, 0], [0, 1]])


@pytest.mark.parametrize(
    "mode",
    [None, "append", "transposed"],
)
def test_pack_unpack_round_trip(mode, example_bool_array):
    if mode == "append":
        bool_array = np.concatenate([example_bool_array, np.ones((4, 1))], axis=1)
        reference_array = bool_array
        samples = bool_array.shape[1]
    elif mode == "transposed":
        bool_array = example_bool_array.T
        reference_array = bool_array.T
        samples = bool_array.shape[0]
    else:
        bool_array = example_bool_array
        reference_array = bool_array
        samples = bool_array.shape[1]
    packed_bytes = pack_bool_array_to_bytes(bool_array)
    unpacked_bool_array = unpack_bytes_to_bool_array(packed_bytes, samples)

    assert np.array_equal(reference_array, unpacked_bool_array)


def test_invalid_rows():
    invalid_bool_array = np.array([[0, 1], [1, 0], [1, 0]])
    with pytest.raises(
        ValueError, match="Number of rows must be four for packing into bytes"
    ):
        pack_bool_array_to_bytes(invalid_bool_array)


def test_pack(example_bool_array):
    packed_bytes = b"i"  # Corresponds to [0, 1, 1, 0, 1, 0, 0, 1]
    assert pack_bool_array_to_bytes(example_bool_array) == packed_bytes


def test_unpack(example_bool_array):
    packed_bytes = b"i"  # Corresponds to [0, 1, 1, 0, 1, 0, 0, 1]
    num_samples = 2
    unpacked_bool_array = unpack_bytes_to_bool_array(packed_bytes, num_samples)

    assert np.array_equal(unpacked_bool_array, example_bool_array)



import io
import pathlib
import re
from datetime import datetime

import numpy as np
import pytest

from RsWaveform import RsWaveform, normalize
from RsWaveform.meta import Meta
from RsWaveform.parent_storage import ParentStorage
from RsWaveform.storage import Storage
from RsWaveform.wv.load import Load
from RsWaveform.wv.save import Save


@pytest.fixture
def reference() -> np.ndarray:
    return np.array([0.2, 0.6]) + 1j * np.array([0.4, 0.8])


@pytest.fixture
def reference_mwv() -> np.ndarray:
    return np.zeros(1000) + 1j * np.zeros(1000)


@pytest.fixture
def reference_huge() -> np.ndarray:
    return np.concatenate((np.array([0.2 + 0.6j] * 5), np.array([0.5 + 0.5j] * 5)))


@pytest.fixture
def meta() -> Meta:
    return Meta(
        **{  # type: ignore[arg-type]
            "type": "SMU-WV",
            "copyright": "Rohde & Schwarz",
            "comment": "Test waveform file",
            "control_length": 2,
            "date": datetime(2023, 1, 5, 10, 3, 52),
            "clock": 100000000.0,
            "marker": {"marker_list_1": [[0, 1], [32, 0], [63, 0]]},
            "control_list": np.array([[0, 1], [1, 0], [1, 0], [0, 1]]),
            "encryption_flag": False,
            "rms": 2.220459,
            "peak": 0.0,
            "samples": 2,
            "reflevel": -20.0,
        }
    )


@pytest.fixture
def meta_mwv() -> Meta:
    return Meta(
        **{  # type: ignore[arg-type]
            "type": "SMU-MWV",
            "copyright": "Rohde&Schwarz",
            "comment": "Test waveform file",
            "date": datetime(2023, 1, 4, 15, 48, 50),
            "clock": 2e8,
            "encryption_flag": False,
            "rms": 200.0,
            "peak": 200.0,
            "samples": 2000,
            "reflevel": -10.0,
        }
    )


# Test Load/Save interfaces #


def test_get_regex_token():
    tokens = Load._get_regex_tokens()
    assert isinstance(tokens, list)
    for x in tokens:
        assert isinstance(x, tuple)


def test_is_waveform_encrypted():
    assert Load._is_waveform_encrypted(b"WWAVEFORM")
    assert not Load._is_waveform_encrypted(b"WAVEFORM")
    with pytest.raises(ValueError):
        Load._is_waveform_encrypted(b"BWAVEFORM")
    with pytest.raises(ValueError):
        assert Load._is_waveform_encrypted(b"UWAVEFORM")


def test_loader(reference_waveform_file_name, meta, reference):
    loader = Load()
    parent_storage = loader.load(reference_waveform_file_name)
    assert len(parent_storage.storages) == 1
    data = parent_storage.storages[0].data
    assert data == pytest.approx(reference, rel=0.001)
    real_valued = np.real(data).astype(np.float16)
    assert real_valued == pytest.approx(
        np.real(reference).astype(np.float16), rel=0.00000001
    )
    imag_valued = np.imag(data).astype(np.float16)
    assert imag_valued == pytest.approx(
        np.imag(reference).astype(np.float16), rel=0.00000001
    )
    obtained_meta = parent_storage.storages[0].meta.copy()
    obtained_control_list = obtained_meta.pop("control_list", np.array([]))
    ref_meta = meta.copy()
    ref_control_list = ref_meta.pop("control_list", np.array([]))
    assert np.array_equal(obtained_control_list, ref_control_list)
    assert obtained_meta == ref_meta


def test_loader_chunks(
    reference_waveform_file_name_huge: str, reference_huge: np.ndarray
):
    loader = Load()
    parent_storage = loader.load_in_chunks(reference_waveform_file_name_huge, 10, 5)
    assert len(parent_storage.storages) == 1
    assert np.real(parent_storage.storages[0].data).astype(np.float16) == pytest.approx(
        np.real(reference_huge).astype(np.float16), rel=0.00000001
    )
    assert np.imag(parent_storage.storages[0].data).astype(np.float16) == pytest.approx(
        np.imag(reference_huge).astype(np.float16), rel=0.00000001
    )


def test_loader_meta_only(reference_waveform_file_name: str, meta):
    loader = Load()
    parent_storage = loader.load_meta(reference_waveform_file_name)
    meta.pop("encryption_flag")
    obtained_meta = parent_storage.storages[0].meta.copy()
    obtained_control_list = obtained_meta.pop("control_list", np.array([]))
    ref_meta = meta.copy()
    ref_control_list = ref_meta.pop("control_list", np.array([]))
    assert np.array_equal(obtained_control_list, ref_control_list)
    assert obtained_meta == ref_meta


def test_loader_mwv(
    reference_waveform_mwv_file_name: str, meta_mwv: dict, reference_mwv: np.ndarray
):
    loader = Load()
    parent_storage = loader.load(reference_waveform_mwv_file_name)
    assert len(parent_storage.storages) == 2
    assert parent_storage.storages[0].data == pytest.approx(reference_mwv, rel=0.001)
    assert np.real(parent_storage.storages[0].data).astype(np.float16) == pytest.approx(
        np.real(reference_mwv).astype(np.float16), rel=0.00000001
    )
    assert np.imag(parent_storage.storages[0].data).astype(np.float16) == pytest.approx(
        np.imag(reference_mwv).astype(np.float16), rel=0.00000001
    )
    assert parent_storage.storages[0].meta == meta_mwv
    assert parent_storage.storages[1].meta == meta_mwv


# Test Save class #
@pytest.mark.parametrize("calculate", [True, False])
def test_saver(
    calculate, reference_waveform_file_name: str, reference: np.ndarray, meta: Meta
):
    np.random.seed(900)
    saver = Save()
    parent_storage = ParentStorage()
    parent_storage.storages[0].data = normalize(reference)
    parent_storage.storages[0].meta = meta
    if calculate:
        parent_storage.storages[0].meta._items["rms"] = None
        parent_storage.storages[0].meta._items["peak"] = None
    with io.BytesIO() as file:
        saver.save(file, parent_storage)
        bins = file.getvalue()
    bins = re.sub(rb"\{DATE:[\d\-;\:]+\}", b"{DATE:2023-01-05;10:03:52}", bins)
    with open(reference_waveform_file_name, "rb") as file:
        ref = file.read()
    ref = re.sub(rb": #", b":#", ref)
    assert bins == ref


def test_saver_mwv(
    reference_waveform_mwv_file_name: str, reference_mwv: np.ndarray, meta_mwv: Meta
):
    np.random.seed(900)

    saver = Save()
    parent_storage = ParentStorage()

    data1 = Storage()
    data1.data = normalize(reference_mwv)
    data1.meta = meta_mwv

    data2 = Storage()
    data2.data = normalize(reference_mwv)
    data2.meta = meta_mwv

    parent_storage.storages = [data1, data2]
    with io.BytesIO() as file:
        saver.save(file, parent_storage)
        bins = file.getvalue()
    bins = re.sub(rb"\{DATE:[\d\-;\:]+\}", b"{DATE:2023-01-04;15:48:50}", bins)
    with open(reference_waveform_mwv_file_name, "rb") as file:
        ref = file.read()
    assert bins == ref


# Test RsWaveform


def test_load(reference_waveform_file_name, meta, reference):
    wv = RsWaveform(file=reference_waveform_file_name)
    assert wv.parent_storage.number_of_storages() == 1
    assert wv.filename == reference_waveform_file_name
    assert wv.data[0] == pytest.approx(reference, rel=0.001)
    assert np.real(wv.data[0]).astype(np.float16) == pytest.approx(
        np.real(reference).astype(np.float16), rel=0.00000001
    )
    assert np.imag(wv.data[0]).astype(np.float16) == pytest.approx(
        np.imag(reference).astype(np.float16), rel=0.00000001
    )
    obtained_meta = wv.meta[0].copy()
    obtained_control_list = obtained_meta.pop("control_list", np.array([]))
    ref_meta = meta.copy()
    ref_control_list = ref_meta.pop("control_list", np.array([]))
    assert np.array_equal(obtained_control_list, ref_control_list)
    assert obtained_meta == ref_meta


def test_load_with_sample_sanity_checking(
    reference_waveform_file_name, tmp_path, caplog
):
    ref_path = pathlib.Path(reference_waveform_file_name)
    mod_file = tmp_path / ref_path.name
    with open(reference_waveform_file_name, "rb") as frp, open(mod_file, "wb") as fwp:
        read_content = frp.read()
        new_content = read_content[:-1] + b"ffff" + read_content[-1:]
        fwp.write(new_content)
    with pytest.raises(
        ValueError,
        match=(
            r"Could not extract [\w]+ data. Malformed [\w]+ section because "
            r"there is no '}' after [\d]+ binary samples."
        ),
    ):
        RsWaveform(file=mod_file)


def test_load_chunks(
    reference_waveform_file_name_huge: str, reference_huge: np.ndarray
):
    wv = RsWaveform(
        file=reference_waveform_file_name_huge, nr_samples=10, samples_offset=5
    )
    assert wv.parent_storage.number_of_storages() == 1
    assert wv.filename == reference_waveform_file_name_huge
    assert np.real(wv.data[0]).astype(np.float16) == pytest.approx(
        np.real(reference_huge).astype(np.float16), rel=0.00000001
    )
    assert np.imag(wv.data[0]).astype(np.float16) == pytest.approx(
        np.imag(reference_huge).astype(np.float16), rel=0.00000001
    )


def test_load_meta_only(reference_waveform_file_name: str, meta: dict):
    wv = RsWaveform(file=reference_waveform_file_name, only_meta_data=True)
    meta.pop("encryption_flag")
    obtained_meta = wv.meta[0].copy()
    obtained_control_list = obtained_meta.pop("control_list", np.array([]))
    ref_meta = meta.copy()
    ref_control_list = ref_meta.pop("control_list", np.array([]))
    assert np.array_equal(obtained_control_list, ref_control_list)
    assert obtained_meta == ref_meta


def test_save(reference_waveform_file_name: str, reference: np.ndarray, meta: dict):
    np.random.seed(900)
    wv = RsWaveform()
    wv.data[0] = normalize(reference)
    wv.meta[0] = meta
    with io.BytesIO() as file:
        wv.save(file)
        bins = file.getvalue()
    bins = re.sub(rb"\{DATE:[\d\-;\:]+\}", b"{DATE:2023-01-05;10:03:52}", bins)
    with open(reference_waveform_file_name, "rb") as file:
        ref = file.read()
    ref = re.sub(rb": #", b":#", ref)
    assert bins == ref



"""Save implementation for generic waveform."""

from __future__ import annotations

import math
import typing

import numpy as np

from ..meta.defaults import META_WV_DEFAULTS as META_DEFAULTS
from ..parent_storage import ParentStorage
from ..save_interface import SaveInterface
from ..storage import Storage
from ..utility.dsp import calculate_peak, calculate_rms
from ..utility.fake_jit import jit
from ..utility.file_handling import write_file_handle
from .utility.array_to_bytes import pack_bool_array_to_bytes

if typing.TYPE_CHECKING:
    from pathlib import Path


class Save(SaveInterface):
    """Save class for generic waveform."""

    def save(
        self,
        file: typing.Union[str, typing.IO, Path],
        datas: ParentStorage,
        scale: float = np.power(2, np.iinfo(np.int16).bits - 1),
    ) -> None:
        """Save waveform data to file."""
        if datas.number_of_storages() == 1:
            self._write(file, datas, scale)
        else:
            self._write_mwv(file, datas, scale)

    def _write(
        self,
        file: typing.Union[str, typing.IO, Path],
        data: ParentStorage,
        scale: float = np.power(2, np.iinfo(np.int16).bits - 1),
    ):
        int16_data = self._prepare_data(data.storages[0], scale)
        with write_file_handle(file) as fp:
            self._write_type(fp, data.storages[0])
            self._write_copyright(fp, data.storages[0])
            self._write_comment(fp, data.storages[0])
            self._write_level_offset(fp, data.storages[0])
            self._write_date(fp, data)
            self._write_clock(fp, data.storages[0])
            self._write_samples(fp, data.storages[0])
            self._write_reflevel(fp, data.storages[0])

            self._write_control_length(fp, data.storages[0])
            self._write_control_list(fp, data.storages[0])
            self._write_marker(fp, data.storages[0])

            self._write_empty_tag(fp)
            self._write_waveform(
                fp, int16_data, data.storages[0].meta.get("encryption_flag", False)
            )

    def _write_mwv(
        self,
        file: typing.Union[str, typing.IO, Path],
        datas: ParentStorage,
        scale: float = np.power(2, np.iinfo(np.int16).bits - 1),
    ):
        tmp_storage = Storage()
        tmp_storage.data = np.array([])
        for d in datas.storages:
            tmp_storage.data = np.append(d.data, tmp_storage.data)
        # first storage used as reference
        tmp_storage.meta.update(datas.storages[0].meta)
        tmp_storage.meta["type"] = "SMU-MWV"
        int16_data = self._prepare_data(tmp_storage, scale)
        with write_file_handle(file) as fp:
            self._write_type(fp, tmp_storage)
            self._write_copyright(fp, tmp_storage)
            self._write_date(fp, datas)
            self._write_samples(fp, tmp_storage)
            self._write_reflevel(fp, tmp_storage)

            self._write_mwv_segment_count(fp, datas)
            self._write_mwv_segment_length(fp, datas.storages)
            self._write_mwv_segment_start(fp, datas.storages)
            self._write_mwv_clock_mode(fp)
            self._write_mwv_level_mode(fp)

            self._write_mwv_clock(fp, datas.storages)
            self._write_mwv_segment_level_offs(fp, datas.storages)
            self._write_mwv_segment_comment(fp, datas.storages)
            self._write_mwv_segment_files(fp, datas.storages)

            self._write_empty_tag(fp)
            self._write_waveform(
                fp, int16_data, tmp_storage.meta.get("encryption", False)
            )

    @staticmethod
    @jit(forceobj=True)
    def _prepare_data(
        data: Storage, scale: float = np.power(2, np.iinfo(np.int16).bits - 1)
    ) -> np.ndarray:
        samples = len(data.data)
        dt = np.int16
        info = np.iinfo(dt)
        int16_samples = np.zeros((2 * samples,)).astype(dt)
        real = np.round(scale * np.real(data.data))
        real[real > info.max] = info.max
        real[real < info.min] = info.min
        int16_samples[slice(0, None, 2)] = real.astype(dt)
        imag = np.round(scale * np.imag(data.data))
        imag[imag > info.max] = info.max
        imag[imag < info.min] = info.min
        int16_samples[slice(1, None, 2)] = imag.astype(dt)
        return int16_samples

    @staticmethod
    def _write_type(file: typing.IO, data: Storage):
        type_data = data.meta.get("type", META_DEFAULTS["type"])
        if isinstance(type_data, tuple):
            type_data = type_data[0]
        line = f"{{TYPE:{type_data}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_copyright(file: typing.IO, data: Storage):
        line = f"{{COPYRIGHT:{data.meta.get('copyright', META_DEFAULTS['copyright'])}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_comment(file: typing.IO, data: Storage):
        line = f"{{COMMENT:{data.meta.get('comment', META_DEFAULTS['comment'])}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_date(file: typing.IO, data: ParentStorage):
        date = data.timestamp
        line = f"{{DATE:{date.strftime('%Y-%m-%d;%H:%M:%S')}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_level_offset(file: typing.IO, data: Storage):
        def invert_values(value: float) -> float:
            if value != 0.0:
                value = value * -1
            return value

        peak = data.meta.get("peak")
        rms = data.meta.get("rms")
        if peak is None:
            peak = invert_values(calculate_peak(data.data))
        if rms is None:
            rms = invert_values(calculate_rms(data.data))
        line = f"{{LEVEL OFFS:{rms:.6f},{peak:.6f}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_clock(file: typing.IO, data: Storage):
        clock = data.meta.get("clock")
        if not clock:
            raise ValueError("Clock is a mandatory parameter!")
        line = f"{{CLOCK:{clock}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_samples(file: typing.IO, data: Storage):
        line = f"{{SAMPLES:{len(data.data)}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_reflevel(file: typing.IO, data: Storage):
        reflevel = data.meta.get("reflevel")
        if reflevel:
            line = f"{{REFLEVEL:{reflevel:.6f}}}"
            file.write(line.encode("utf-8"))

    @staticmethod
    def _write_marker(file: typing.IO, data: Storage):
        marker_keys = [
            key for key in data.meta.get("marker", {}) if key.startswith("marker_list")
        ]
        if not marker_keys:
            return
        for marker in marker_keys:
            marker_string = ";".join(
                [
                    ":".join([str(el) for el in entry])
                    for entry in data.meta["marker"][marker]
                ]
            )
            line = f"{{{marker.replace('_', ' ').upper()}: {marker_string}}}"
            file.write(line.encode("utf-8"))

    @staticmethod
    def _write_control_length(file: typing.IO, data: Storage):
        control_length = data.meta.get("control_length")
        if not control_length:
            return
        line = f"{{CONTROL LENGTH:{control_length}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_control_list(file: typing.IO, data: Storage):
        control_list = data.meta.get("control_list")
        if control_list is None:
            return
        if isinstance(control_list, list):
            control_list = np.array(control_list)
        list_bytes = pack_bool_array_to_bytes(control_list)
        line = f"{{CONTROL LIST WIDTH4-{len(list_bytes) + 1}:#"
        file.write(line.encode())
        file.write(list_bytes)
        file.write(b"}")

    @staticmethod
    def _write_empty_tag(file: typing.IO):
        scale = 512
        rand_empty = int(np.around(np.random.rand() * scale))
        line = f"{{EMPTYTAG-{rand_empty + 1}:#"
        file.write(line.encode("utf-8"))
        empty = 32 * np.ones((rand_empty,)).astype(np.uint8)
        file.write(empty.tobytes())
        file.write(b"}")

    @staticmethod
    def _write_waveform(
        file: typing.IO, int16_data: np.ndarray, encryption_flag: bool = False
    ):
        prefix = "W" if encryption_flag else ""
        line = f"{{{prefix}WAVEFORM-{len(int16_data) * 2 + 1}:#"
        file.write(line.encode())
        chunk_size = 82000000  # randomly chosen number. Might be changed!
        number_blocks = math.ceil(int16_data.size / chunk_size)
        for block_idx in range(number_blocks):
            idx_start = block_idx * chunk_size
            idx_stop = (block_idx + 1) * chunk_size
            if idx_stop > int16_data.size - 1:
                idx_stop = int16_data.size
            bins = int16_data[idx_start:idx_stop].tobytes()
            file.write(bins)
        file.write(b"}")

    @staticmethod
    def _write_mwv_segment_count(file: typing.IO, datas: ParentStorage):
        line = f"{{MWV_SEGMENT_COUNT:{datas.number_of_storages()}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_mwv_segment_length(file: typing.IO, datas: typing.List[Storage]):
        segment_length = []
        for data in datas:
            segment_length.append(str(len(data.data)))
        line = f"{{MWV_SEGMENT_LENGTH:{','.join(segment_length)}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_mwv_segment_start(file: typing.IO, datas: typing.List[Storage]):
        segment_start = []
        start_segment = 0
        for data in datas:
            segment_start.append(str(start_segment))
            start_segment += len(data.data)
        line = f"{{MWV_SEGMENT_START:{','.join(segment_start)}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_mwv_clock(file: typing.IO, datas: typing.List[Storage]):
        max_clock: float = 0.0
        all_clocks = []
        for data in datas:
            if float(data.meta.get("clock", 0)) > max_clock:
                max_clock = float(data.meta.get("clock", 0))
            all_clocks.append(str(data.meta.get("clock")))
        line = f"{{CLOCK:{max_clock}}}"
        file.write(line.encode("utf-8"))
        line = f"{{MWV_SEGMENT_CLOCK:{','.join(all_clocks)}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_mwv_segment_level_offs(file: typing.IO, datas: typing.List[Storage]):
        level_offs = []
        for data in datas:
            level_offs.append(str(data.meta.get("rms", 0)))
            level_offs.append(str(data.meta.get("peak", 0)))
        line = f"{{MWV_SEGMENT_LEVEL_OFFS:{','.join(level_offs)}}}"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_mwv_clock_mode(file: typing.IO, mode: str = "UNCHANGED"):
        line = f"{{MWV_SEGMENT_CLOCK_MODE:{mode}}}"
        file.write(line.encode())

    @staticmethod
    def _write_mwv_level_mode(file: typing.IO, mode: str = "UNCHANGED"):
        line = f"{{MWV_SEGMENT_LEVEL_MODE:{mode}}}"
        file.write(line.encode())

    @staticmethod
    def _write_mwv_segment_comment(file: typing.IO, datas: typing.List[Storage]):
        for index, data in enumerate(datas):
            line = f"{{MWV_SEGMENT{index}_COMMENT:{data.meta.get('comment')}}}"
            file.write(line.encode("utf-8"))

    @staticmethod
    def _write_mwv_segment_files(file: typing.IO, datas: typing.List[Storage]):
        file_names = []
        for data in datas:
            if "filename" in data.meta:
                file_names.append(str(data.meta.get("filename", "")))
        if file_names:
            line = f"{{MWV_SEGMENT_FILES:{','.join(file_names)}}}"
            file.write(line.encode("utf-8"))



"""Storage wrapper class for list of Storages."""

import abc
import typing
from datetime import datetime

from .storage import Storage


class ParentStorage(abc.ABC):
    """ParentStorage class implementation.

    Args:
        number_of_storages (int, optional): Initialize number of storages.
            Defaults to 1.
        no_defaults (bool, optional): Storage meta data should have no defaults at
            initialization. Defaults to False.

    Raises:
        ValueError: If number of storages is smaller than 1.
    """

    def __init__(self, number_of_storages: int = 1, no_defaults: bool = False):
        """Initialize ParentStorage class."""
        if number_of_storages <= 0:
            raise ValueError("Only positive number of storages allowed!")
        self._storages = [
            Storage(no_defaults=no_defaults) for _ in range(number_of_storages)
        ]
        self._filename: str = ""
        self._timestamp: datetime = datetime.now()

    @property
    def timestamp(self) -> datetime:
        """Get timestamp."""
        return self._timestamp

    def number_of_storages(self) -> int:
        """Get number of storages."""
        return len(self._storages)

    def add_storage(self, storage: Storage):
        """Add storage to parentStorage."""
        self._storages.append(storage)

    @property
    def storages(self) -> typing.List[Storage]:
        """Get storages."""
        return self._storages

    @storages.setter
    def storages(self, storages: typing.List[Storage]):
        """Set all storages."""
        self._storages = storages

    @property
    def filename(self) -> str:
        """Get filename."""
        return self._filename

    @filename.setter
    def filename(self, filename: str):
        """Set filename."""
        self._filename = filename



"""Load R&S waveform files."""

from __future__ import annotations

import abc
import typing

from .parent_storage import ParentStorage

if typing.TYPE_CHECKING:
    from pathlib import Path


class LoadInterface(abc.ABC):
    """Load R&S waveform implementation."""

    @abc.abstractmethod
    def load(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Abstract load implementation."""

    @abc.abstractmethod
    def load_in_chunks(
        self, file: typing.Union[str, typing.IO, Path], samples: int, offset: int
    ) -> ParentStorage:
        """Abstract load implementation for chunks."""

    @abc.abstractmethod
    def load_meta(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Abstract load implementation for meta information only."""



"""Load implementation for generic waveform."""

from __future__ import annotations

import logging
import re
import typing
from datetime import datetime

import numpy as np

from ..load_interface import LoadInterface
from ..meta import Meta
from ..parent_storage import ParentStorage
from ..storage import Storage
from ..utility.fake_jit import jit
from ..utility.file_handling import read_file_handle
from .utility.array_to_bytes import unpack_bytes_to_bool_array

if typing.TYPE_CHECKING:
    from pathlib import Path


LOGGER = logging.getLogger(__name__)


class Load(LoadInterface):
    """Load class for generic waveform."""

    def load(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load waveform data from file."""
        with read_file_handle(file) as fp:
            content = fp.read()

        tags = self._split_data_via_tags_meta(content)
        samples = int(tags["samples"])
        mwv_segment_count = int(tags.pop("mwv_segment_count", 1))

        tags.update(self._split_data_via_tags_waveform(content))
        tags.update(self._split_data_via_tags_control_list_width4(content, samples))

        parent_storage = ParentStorage()
        if isinstance(file, str):
            parent_storage.filename = file
        storages = [Storage() for _ in range(mwv_segment_count)]
        if mwv_segment_count > 1:  # multi segment waveform
            value = tags.pop("mwv_segment_length").decode()
            tags = self._extract_mwv_comments(mwv_segment_count, tags, content)
            segment_lengths = [int(length) for length in value.split(",")]

        else:
            segment_lengths = [samples]

        iq_data = self._extract_iq(tags)
        for index in range(mwv_segment_count):
            start_index = sum(segment_lengths[0:index])
            end_index = segment_lengths[index] + start_index
            storages[index].data = iq_data[start_index:end_index]
            meta = self._extract_meta(tags, index)
            meta = self._handle_mwv_meta_data(meta, index)
            storages[index].meta = Meta(**meta)
        if samples != len(iq_data):
            LOGGER.warning(
                (
                    "Sanity problem. Sample count from tag '%s' and actual sample "
                    "count from bytes '%s' does not match."
                ),
                samples,
                len(iq_data),
            )
        parent_storage.storages = storages
        return parent_storage

    @staticmethod
    def _read_chunks(stream: typing.IO, separators: list) -> typing.Tuple[bytes, bytes]:
        regex = "(" + "|".join(separators) + ")"
        chunk_size = 4096
        buffer = b""
        while True:
            chunk = stream.read(chunk_size)
            if not chunk:
                return buffer, b""
            buffer += chunk
            while True:
                try:
                    part, separator, buffer = re.split(regex.encode("utf-8"), buffer)
                except ValueError:
                    break
                else:
                    return part, separator + buffer

    def load_in_chunks(
        self,
        file: typing.Union[str, typing.IO, Path],
        samples: int,
        offset: int,
    ) -> ParentStorage:
        """Load chunk waveform data from file."""
        separators = ["{WWAVEFORM", "{WAVEFORM"]
        with read_file_handle(file) as fp:
            header_content, content = self._read_chunks(fp, separators)
            content += fp.read((samples + 4) * 4 + 100)
            # +4 due to opt words and 100 just an offset to make sure we get everything

        tags = self._split_data_via_tags_meta(header_content)
        tags.update(self._split_data_via_tags_waveform(content, True))
        tags.update(
            self._split_data_via_tags_control_list_width4(content, samples, True)
        )

        mwv_segment_count = int(tags.pop("mwv segment count", 1))
        if mwv_segment_count > 1:  # multi segment waveform
            raise ValueError("feature only supported for non multi segment wv")
        parent_storage = ParentStorage()
        if isinstance(file, str):
            parent_storage.filename = file
        storages = [Storage() for _ in range(mwv_segment_count)]
        iq_data = self._extract_iq_chunks(tags, samples, offset)
        index = 0
        storages[index].data = iq_data
        meta = self._extract_meta(tags, index)
        meta = self._handle_mwv_meta_data(meta, index)
        storages[index].meta = Meta(**meta)
        if samples != len(iq_data):
            raise ValueError(
                "Sanity problem. Sample setting and actual sample count does not match."
            )
        parent_storage.storages = storages
        return parent_storage

    def load_meta(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load meta information only from wv file."""
        separators = ["{WWAVEFORM", "{WAVEFORM"]
        with read_file_handle(file) as fp:
            header_content, _ = self._read_chunks(fp, separators)

        tags = self._split_data_via_tags_meta(header_content)
        samples = int(tags["samples"])
        tags.update(
            self._split_data_via_tags_control_list_width4(header_content, samples)
        )
        meta = self._extract_meta(tags)
        parent_storage = ParentStorage()
        parent_storage.storages[0].meta = Meta(**meta)
        parent_storage.storages[0].data = np.array([])
        return parent_storage

    @staticmethod
    @jit(forceobj=True)
    def _extract_iq(tags: dict) -> np.ndarray:
        iq_bytes = tags.pop("waveform")
        if not iq_bytes:
            raise ValueError("No waveform data availabe.")
        from_dtype = np.dtype(np.int16)
        to_dtype = np.dtype(np.float16)
        i_values = Load._fix_to_double(
            np.frombuffer(
                buffer=Load._extract_pairs(iq_bytes, 0),
                dtype=from_dtype,
            ),
            to_dtype,
        )
        q_values = Load._fix_to_double(
            np.frombuffer(
                buffer=Load._extract_pairs(iq_bytes, 1),
                dtype=from_dtype,
            ),
            to_dtype,
        )
        iq = i_values + 1j * q_values
        return iq

    @staticmethod
    @jit(forceobj=True)
    def _extract_iq_chunks(
        tags: dict, nr_samples: int, samples_offset: int
    ) -> np.ndarray:
        iq_bytes = tags.pop("waveform")

        if not iq_bytes:
            raise ValueError("No waveform data availabe.")
        from_dtype = np.dtype(np.int16)
        to_dtype = np.dtype(np.float16)
        i_values = Load._fix_to_double(
            np.frombuffer(
                buffer=Load._extract_pairs(
                    iq_bytes, 0, length=nr_samples, offset=samples_offset
                ),
                dtype=from_dtype,
            ),
            to_dtype,
        )
        q_values = Load._fix_to_double(
            np.frombuffer(
                buffer=Load._extract_pairs(
                    iq_bytes, 1, length=nr_samples, offset=samples_offset
                ),
                dtype=from_dtype,
            ),
            to_dtype,
        )
        iq = i_values + 1j * q_values
        return iq

    @staticmethod
    @jit(forceobj=True)
    def _fix_to_double(values: np.ndarray, to_float_type: np.dtype) -> np.ndarray:
        dt_info = np.iinfo(values.dtype)
        maximum = to_float_type.type(dt_info.max)
        converted = to_float_type.type(values) / maximum
        return converted

    @staticmethod
    @jit(forceobj=True)
    def _extract_pairs(
        buf: bytes, start: int = 0, offset: int = 0, length: int = 0
    ) -> bytes:
        """Extract data pairs from binary buffer."""
        actual_length = length * 4
        actual_offset = offset * 4
        single_pair = bytearray()
        if actual_length == 0 or actual_length > len(buf):
            actual_length = len(buf)
            actual_offset = 0
        if actual_length + actual_offset > len(buf):
            actual_offset = len(buf) - actual_length
        indices = np.arange(start * 2 + actual_offset, actual_offset + actual_length, 4)
        for idx in indices:
            temp = buf[idx : idx + 2]
            single_pair.extend(temp)
        return single_pair

    @staticmethod
    def _extract_meta(tags: dict, index: int = 0) -> dict:
        meta: typing.Dict[str, typing.Any] = {}
        marker = {}
        for key, value in tags.items():
            if isinstance(value, bytes):
                value = value.decode()
            if key == "type":
                value = tuple(x.strip() for x in value.split(","))[0]
            elif key == "level_offs":
                offset, peak = value.split(",")
                value = None
                meta.update(rms=float(offset), peak=float(peak))
            elif key in ["vector_max", "clock"]:
                value = float(value)
            elif key == "date":
                value = datetime.strptime(value, "%Y-%m-%d;%H:%M:%S")
            elif key == "control_length":
                value = int(value)
            elif re.match(r"marker_list_\d+", key):
                marker_list = []
                for sub in value.split(";"):
                    marker_list.append([int(x) for x in sub.split(":")])
                if marker_list:
                    marker.update({key: marker_list})
                continue
            elif key == "samples":
                value = int(value)
            elif key == "reflevel":
                value = float(value)
            elif key == "mwv_segment_level_offs":
                possible_values = value.split(",")
                offset = possible_values[index * 2]
                peak = possible_values[index * 2 + 1]
                value = (float(offset), float(peak))
            elif key == "mwv_segment_clock":
                value = float(value.split(",")[index])
            elif key == "mwv_segment_length":
                value = float(value.split(",")[index])
            elif (
                re.search(r"mwv_segment\d{1,}_comment", key)
                and key != f"mwv_segment{index}_comment"
            ):
                continue
            if value is None:
                continue
            meta.update({key: value})
        meta.update(marker=marker)
        return meta

    @staticmethod
    def _handle_mwv_meta_data(meta: dict, index: int = 0) -> dict:
        if "mwv_segment_clock" in meta:
            meta["clock"] = meta.pop("mwv_segment_clock")
        if "mwv_segment_level_offs" in meta:
            offset, peak = meta.pop("mwv_segment_level_offs")
            meta.update(rms=float(offset), peak=float(peak))
        if "mwv_segment_length" in meta:
            meta["samples"] = meta.pop("mwv_segment_length")
        if f"mwv_segment{index}_comment" in meta:
            meta["comment"] = meta.pop(f"mwv_segment{index}_comment")
        return meta

    @staticmethod
    @jit(forceobj=True)
    def _split_data_via_tags_meta(content: bytes) -> dict:
        tok_regex = Load._create_regex_pattern()
        extracted: typing.Dict[str, typing.Any] = {}
        for mo in re.finditer(tok_regex, content):
            if not mo.lastgroup:
                continue
            kind = mo.lastgroup
            value = mo.group()
            extracted.update({kind: value})
        return extracted

    @staticmethod
    def _split_data_via_tags_waveform(
        content: bytes, partial_read: bool = False
    ) -> dict:
        extracted: typing.Dict[str, typing.Any] = {}
        # search for waveform data
        waveform, encryption_flag = Load._extract_waveform(content, partial_read)
        extracted.update({"waveform": waveform, "encryption_flag": encryption_flag})
        return extracted

    @staticmethod
    def _split_data_via_tags_control_list_width4(
        content: bytes, samples: int, partial_read: bool = False
    ) -> dict:
        extracted: typing.Dict[str, typing.Any] = {}
        # search for waveform data
        control_list = Load._extract_binary_tags(
            content, "CONTROL LIST WIDTH4", partial_read
        )
        if control_list:
            unpacked_list = unpack_bytes_to_bool_array(control_list, samples)
            extracted.update({"control_list": unpacked_list})
        return extracted

    @staticmethod
    @jit(forceobj=True)
    def _extract_waveform(
        content: bytes, partial_read: bool = False
    ) -> typing.Tuple[bytes, bool]:
        m_waveform = re.search(rb"(?<={)(\w{0,1}WAVEFORM)", content)
        if not m_waveform:
            raise ValueError("Waveform does not contain waveform tag.")
        encryption_flag = Load._is_waveform_encrypted(
            content[m_waveform.start() : m_waveform.end()]
        )
        iq = Load._extract_binary_tags(
            content, f"{'W' if encryption_flag else ''}WAVEFORM", partial_read
        )
        return iq, encryption_flag

    @staticmethod
    @jit(forceobj=True)
    def _extract_binary_tags(
        content: bytes, tag: str, partial_read: bool = False
    ) -> typing.Optional[bytes]:
        m_tag = re.search(rb"(?<={)(%s)" % tag.encode("utf-8"), content)
        if not m_tag:
            return None
        start_index = m_tag.start()
        tag_content = content[start_index:]
        m_content = re.findall(
            rb"(?<=%s\-)([\d]+)(?=\:)" % tag.encode("utf-8"), tag_content
        )
        if not m_content:
            raise ValueError(f"Waveform does not contain {tag} byte count.")
        count = int(m_content[0].decode())
        pattern = rf"{count}:( |)#".encode("utf-8")  # Binary data
        m_pattern = re.search(pattern, tag_content)
        if not m_pattern:
            raise ValueError(
                f"Could not extract {tag} data. Could not find {tag} section."
            )
        start_index = m_pattern.end()
        end_index = (len(content) if partial_read else start_index + count) - 1
        if start_index + count > len(tag_content) and partial_read is False:
            raise ValueError(
                f"Could not extract {tag} data. Provided byte count inconclusive."
            )
        binary = tag_content[start_index:end_index]
        if partial_read is False:
            closing = tag_content[end_index]
            if chr(closing) != "}":
                raise ValueError(
                    f"Could not extract {tag} data. Malformed {tag} section because "
                    f"there is no '}}' after {count} binary samples."
                )

        return binary

    @staticmethod
    @jit(forceobj=True)
    def _is_waveform_encrypted(waveform_tag: bytes) -> bool:
        encrypted = re.findall(rb"(\w{0,1})WAVEFORM", waveform_tag, re.IGNORECASE)
        if encrypted[0] == b"":
            return False
        if encrypted[0] != b"W":
            raise ValueError(
                f"Not supported waveform tag format {waveform_tag.decode()}"
            )
        return True

    @staticmethod
    def _create_regex_pattern() -> bytes:
        token_specification = Load._get_regex_tokens()
        tok_regex = "|".join(
            rf"(?P<{pair[0].lower().replace(' ', '_').replace('.', '')}"
            rf"{'_%s' % pair[1].replace(' ', '') if pair[1] != '' else ''}>"
            rf"(?<={{{pair[0]}{pair[1]}:)\s{{0,1}}({pair[2]})(?=}}))"
            for pair in token_specification
        )
        return tok_regex.encode("utf-8")

    @staticmethod
    def _get_regex_tokens() -> list:
        token_specification = [
            ("TYPE", r"", r"([\w-]+)(,[\s]*([\d]+))?"),  # Waveform type
            ("COPYRIGHT", r"", r"[\w \&]+"),  # Copyright of waveform
            ("COMMENT", r"", r"[\w \.]+"),  # Copyright of waveform
            ("DATE", r"", r"[\d\-;\:]+"),  # Timestamp of waveform
            ("SAMPLES", r"", r"[\de\-\+]+"),  # Sample count
            ("CLOCK", r"", r"[\d\.e\-\+]+"),  # Waveform sampling frequency
            ("REFLEVEL", r"", r"[\d\.e\-\+]+"),  # Reference level for playback
            ("VECTOR MAX", r"", r"[\d\.\-e\+]+"),  # tbd
            ("LEVEL OFFS", r"", r"([\d\.\-e\+]+),([\d\.\-e\+]+)"),
            # Level offset to dBFS
            ("CONTROL LENGTH", r"", r"[\d\.e\-\+]+"),  # control length count
            ("MWV_SEGMENT_LENGTH", r"", r"[\d,]+"),  # multi segment length
            ("MWV_SEGMENT_COUNT", r"", r"[\de\.\+\-]+"),  # multi segment counter
            ("MWV_SEGMENT_CLOCK", r"", r"[\d\.e\-\+,]+"),
            # list of clocks in multi segment wv
            ("MWV_SEGMENT_LEVEL_OFFS", r"", r"[\d\.e\-\+,]+"),
            # list of level offsets in multi segment wv
        ]
        token_specification.extend(
            [
                ("MARKER LIST", f" {idx}", r"[\d:;]+") for idx in range(1, 5)
            ],  # Marker list mode x
        )

        token_specification.extend(
            [
                ("MARKER MODE", f" {idx}", r"[\w]+") for idx in range(1, 5)
            ],  # Marker list mode x
        )
        return token_specification

    @staticmethod
    def _get_regex_mwv_comment_tokens(mwv_segment_count: int) -> list:
        token_specification = [
            (f"MWV_SEGMENT{idx}_COMMENT", r"[\w \.]+")
            for idx in range(mwv_segment_count)
        ]

        return token_specification

    @staticmethod
    def _create_regex_mwv_comment_pattern(mwv_segment_count: int) -> bytes:
        token_specification = Load._get_regex_mwv_comment_tokens(mwv_segment_count)
        tok_regex = "|".join(
            rf"(?P<{pair[0].lower().replace(' ', '_').replace('.', '')}>"
            rf"(?<={{{pair[0]}:)\s{{0,1}}({pair[1]})(?=}}))"
            for pair in token_specification
        )
        return tok_regex.encode("utf-8")

    # due to the high number of mwv comments, we only extract them, if needed
    # and only the number of comments, which might be available
    @staticmethod
    def _extract_mwv_comments(
        mwv_segment_count: int, tags: dict, content: bytes
    ) -> dict:
        # due to performance issues, in the first step we only grab the part of the
        # content containing the meta, throwing away all samples
        regex = rb"^(.*?)\{W{0,1}WAV"
        for meta_part in re.finditer(regex, content):
            tok_regex = Load._create_regex_mwv_comment_pattern(mwv_segment_count)
            for mo in re.finditer(tok_regex, meta_part.group()):
                if not mo.lastgroup:
                    continue
                kind = mo.lastgroup
                value = mo.group()
                tags.update({kind: value})
        return tags



"""Common DSP methods."""

import typing

import numpy as np

from .fake_jit import jit


@jit(forceobj=True)
def normalize(data: np.ndarray, reference: typing.Optional[float] = None):
    """Normalize data against a reference value."""
    if not reference:
        reference = float(np.float16(1 - np.float_power(2, -15)))
        # by default all waveforms are 16-bit quantized
    if np.max(np.abs(data)) != 0:
        return np.multiply(data, reference / np.max(np.abs(data)))
    else:
        return data


@jit(forceobj=True)
def convert_to_db(value: np.ndarray) -> np.ndarray:
    """Convert value from amplitude to dB."""
    return 20 * np.log10(np.float16(value))


@jit(forceobj=True)
def calculate_peak(data: np.ndarray) -> np.ndarray:
    """Calculate peak value in dB."""
    maximum_value = np.max(np.abs(data))
    return convert_to_db(np.float16(maximum_value))


@jit(forceobj=True)
def calculate_rms(data: np.ndarray) -> np.ndarray:
    """Calculate root-mean-square (RMS) value in dB."""
    magnitude = data * np.conj(data)
    rms = np.sqrt(np.mean(magnitude))
    return convert_to_db(np.abs(rms))


def calculate_par(data) -> float:
    """Calculate peak-to-average-ratio (PAR) value in dB."""
    return calculate_peak(data) - calculate_rms(data)



import io

import numpy as np
import pytest

from RsWaveform import Iqw
from RsWaveform.iqw.load import Load
from RsWaveform.iqw.save import Save
from RsWaveform.meta import Meta
from RsWaveform.parent_storage import ParentStorage


@pytest.fixture
def reference() -> np.ndarray:
    return np.array([0.2, 0.6]) + 1j * np.array([0.4, 0.8])


@pytest.fixture
def reference_huge() -> np.ndarray:
    return np.concatenate((np.array([0.2 + 0.6j] * 5), np.array([0.5 + 0.5j] * 5)))


def test_loader_load(reference_iqw_file_name, reference):
    loader = Load()
    parent_storage = loader.load(reference_iqw_file_name)
    assert parent_storage.number_of_storages() == 1
    assert parent_storage.storages[0].data == pytest.approx(reference, rel=0.0001)
    assert isinstance(parent_storage.storages[0].meta, Meta)


def test_saver_save(reference_iqw_file_name, reference: np.ndarray):
    saver = Save()
    parent_storage = ParentStorage()
    parent_storage.storages[0].data = reference
    with io.BytesIO() as file:
        saver.save(file, parent_storage)
        bins = file.getvalue()
    with open(reference_iqw_file_name, "rb") as file:
        ref = file.read()
    assert bins == ref


def test_loader_chunks(reference_iqw_file_name_huge: str, reference_huge: np.ndarray):
    loader = Load()
    parent_storage = loader.load_in_chunks(reference_iqw_file_name_huge, 10, 5)
    assert len(parent_storage.storages) == 1
    assert np.real(parent_storage.storages[0].data).astype(np.float16) == pytest.approx(
        np.real(reference_huge).astype(np.float16), rel=0.00000001
    )
    assert np.imag(parent_storage.storages[0].data).astype(np.float16) == pytest.approx(
        np.imag(reference_huge).astype(np.float16), rel=0.00000001
    )
    assert isinstance(parent_storage.storages[0].meta, Meta)


def test_loader_only_meta(reference_iqw_file_name: str):
    loader = Load()
    with pytest.raises(Exception):
        loader.load_meta(reference_iqw_file_name)


def test_load(reference_iqw_file_name, reference):
    iqw = Iqw(file=reference_iqw_file_name)
    assert iqw.parent_storage.number_of_storages() == 1
    assert iqw.filename == reference_iqw_file_name
    assert iqw.data[0] == pytest.approx(reference, rel=0.0001)
    assert isinstance(iqw.meta[0], Meta)


def test_load_chunks(reference_iqw_file_name_huge, reference_huge):
    iqw = Iqw(file=reference_iqw_file_name_huge, nr_samples=10, samples_offset=5)
    assert iqw.parent_storage.number_of_storages() == 1
    assert iqw.filename == reference_iqw_file_name_huge
    assert iqw.data[0] == pytest.approx(reference_huge, rel=0.0001)
    assert isinstance(iqw.meta[0], Meta)


def test_save(reference_iqw_file_name: str, reference: np.ndarray):
    np.random.seed(900)
    iqtar = Iqw()
    iqtar.data[0] = reference
    with io.BytesIO() as file:
        iqtar.save(file)
        bins = file.getvalue()
    with open(reference_iqw_file_name, "rb") as file:
        ref = file.read()
    assert bins == ref



import numpy as np

import RsWaveform.utility.dsp as dsp


def get_array() -> np.ndarray:
    values = np.arange(16)
    return values


def test_convert_to_db():
    db = dsp.convert_to_db(10)
    assert db == 20


def test_calculate_peak():
    data = get_array()
    peak = dsp.calculate_peak(data)
    assert peak == dsp.convert_to_db(np.max(data))


def test_calculate_rms():
    data = get_array()
    rms = dsp.calculate_rms(data)
    assert rms == 18.896484375


def test_calculate_par():
    data = get_array()
    peak = dsp.calculate_peak(data)
    rms = dsp.calculate_rms(data)
    par = dsp.calculate_par(data)
    assert par == peak - rms


def test_normalize():
    # Test default reference
    reference = float(np.float16(1 - np.float_power(2, -15)))
    data = get_array()
    normalized = dsp.normalize(data)
    assert np.max(np.abs(normalized)) == reference

    # Test any reference
    reference = 2
    data = get_array()
    normalized = dsp.normalize(data, reference)
    assert np.max(np.abs(normalized)) == reference



import os
from datetime import datetime

import numpy as np
import pytest

from RsWaveform import IqTar
from RsWaveform.iqtar.load import Load
from RsWaveform.iqtar.save import Save
from RsWaveform.meta import Meta
from RsWaveform.parent_storage import ParentStorage
from RsWaveform.storage import Storage


@pytest.fixture
def reference() -> np.ndarray:
    return np.array([0.2, 0.6]) + 1j * np.array([0.4, 0.8])


@pytest.fixture
def reference_huge() -> np.ndarray:
    return np.concatenate((np.array([0.2 + 0.6j] * 5), np.array([0.5 + 0.5j] * 5)))


@pytest.fixture()
def meta():
    _meta = Meta(
        **{  # type: ignore[arg-type]
            "clock": 10000.0,
            "scalingfactor": 1.0,
            "datatype": "float32",
            "format": "complex",
            "name": "Python iq.tar Writer (iqdata.py)",
            "comment": "RS WaveForm, TheAE-RA",
            "date": datetime(2023, 3, 1, 10, 19, 37, 43312),
        }
    )
    return _meta


@pytest.fixture()
def meta_multi_channel():
    _meta = Meta(
        **{  # type: ignore[arg-type]
            "clock": 10000.0,
            "scalingfactor": 1.0,
            "datatype": "float32",
            "format": "complex",
            "name": "Python iq.tar Writer",
            "comment": "RS WaveForm, TheAE-RA",
            "date": datetime(2023, 3, 3, 8, 9, 25, 842264),
        }
    )
    return _meta


def test_loader_load(reference_iqtar_file_name: str, reference: np.ndarray, meta: Meta):
    loader = Load()
    parent_storage = loader.load(reference_iqtar_file_name)
    assert parent_storage.number_of_storages() == 1
    assert parent_storage.storages[0].data == pytest.approx(reference, rel=0.0001)
    assert meta == parent_storage.storages[0].meta


def test_loader_load_multi_channel(
    reference_iqtar_file_name_multi_channel: str,
    reference: np.ndarray,
    meta_multi_channel: dict,
):
    loader = Load()
    parent_storage = loader.load(reference_iqtar_file_name_multi_channel)
    assert parent_storage.number_of_storages() > 1
    assert parent_storage.storages[0].meta == meta_multi_channel


def test_loader_chunk(
    reference_iqtar_file_name_huge: str, reference_huge: np.ndarray, meta: dict
):
    loader = Load()
    parent_storage = loader.load_in_chunks(reference_iqtar_file_name_huge, 10, 5)
    assert parent_storage.number_of_storages() == 1
    assert parent_storage.storages[0].data == pytest.approx(reference_huge, rel=0.0001)
    meta["date"] = datetime(2023, 4, 11, 10, 32, 30, 586434)
    meta["name"] = "Python iq.tar Writer"
    assert meta == parent_storage.storages[0].meta


def test_loader_only_meta(reference_iqtar_file_name: str, meta: dict):
    loader = Load()
    parent_storage = loader.load_meta(reference_iqtar_file_name)
    meta["datafilename"] = "dummy.complex.1ch.float32"
    assert parent_storage.storages[0].meta == meta


def test_saver_save(reference_iqtar_file_name, reference: np.ndarray, meta: Meta):
    saver = Save()
    parent_storage = ParentStorage()
    parent_storage.storages[0].data = reference
    parent_storage.storages[0].meta = meta
    dirname = os.path.dirname(__file__)
    tmp_folder = os.path.join(dirname, "data", "dummy_tmp.iq.tar")
    saver.save(tmp_folder, parent_storage)
    with open(reference_iqtar_file_name, "rb") as file:
        bins = file.read()
    os.remove(tmp_folder)
    with open(reference_iqtar_file_name, "rb") as file:
        ref = file.read()
    assert bins == ref


def test_saver_save_multi_channel(
    reference_iqtar_file_name_multi_channel: str,
    reference: np.ndarray,
    meta_multi_channel: Meta,
):
    saver = Save()
    parent_storage = ParentStorage()
    storage_1 = Storage()
    storage_1.data = reference
    storage_1.meta = meta_multi_channel

    storage_2 = Storage()
    storage_2.data = reference
    storage_2.meta = meta_multi_channel
    parent_storage.storages = [storage_1, storage_2]

    dirname = os.path.dirname(__file__)
    tmp_folder = os.path.join(dirname, "data", "dummy_tmp.iq.tar")
    saver.save(tmp_folder, parent_storage)
    with open(reference_iqtar_file_name_multi_channel, "rb") as file:
        bins = file.read()
    os.remove(tmp_folder)
    with open(reference_iqtar_file_name_multi_channel, "rb") as file:
        ref = file.read()
    assert bins == ref


def test_load(reference_iqtar_file_name: str, reference, meta):
    iqtar = IqTar(file=reference_iqtar_file_name)
    assert iqtar.parent_storage.number_of_storages() == 1
    assert iqtar.filename == reference_iqtar_file_name
    assert iqtar.data[0] == pytest.approx(reference, rel=0.0001)
    assert iqtar.meta[0] == meta


def test_load_chunks(reference_iqtar_file_name_huge: str, reference_huge, meta):
    iqtar = IqTar(file=reference_iqtar_file_name_huge, nr_samples=10, samples_offset=5)
    assert iqtar.parent_storage.number_of_storages() == 1
    assert iqtar.filename == reference_iqtar_file_name_huge
    assert iqtar.data[0] == pytest.approx(reference_huge, rel=0.0001)
    meta["date"] = datetime(2023, 4, 11, 10, 32, 30, 586434)
    meta["name"] = "Python iq.tar Writer"
    assert iqtar.meta[0] == meta


def test_load_only_meta(reference_iqtar_file_name: str, meta: dict):
    iqtar = IqTar(file=reference_iqtar_file_name, only_meta_data=True)
    meta["datafilename"] = "dummy.complex.1ch.float32"
    assert iqtar.meta[0] == meta


def test_save(reference_iqtar_file_name: str, reference: np.ndarray, meta: dict):
    np.random.seed(900)
    iqtar = IqTar()
    iqtar.data[0] = reference
    iqtar.meta[0] = meta
    dirname = os.path.dirname(__file__)
    tmp_folder = os.path.join(dirname, "data", "dummy_tmp.iq.tar")
    iqtar.save(tmp_folder)
    with open(reference_iqtar_file_name, "rb") as file:
        bins = file.read()
    os.remove(tmp_folder)
    with open(reference_iqtar_file_name, "rb") as file:
        ref = file.read()
    assert bins == ref



"""Storage wrapper class for waveform data."""

import abc
import typing

import numpy as np

from .meta import Meta

try:
    msgpack_import = True
    import msgpack as pickle
    import msgpack_numpy as m

    m.patch()

except ImportError:
    msgpack_import = False
    import pickle
from datetime import datetime


def decode_datetime(obj):
    """Decode datetime for msgpack."""
    if "__datetime__" in obj:
        obj = datetime.strptime(obj["as_str"], "%Y%m%dT%H:%M:%S.%f")
    return obj


def encode_datetime(obj):
    """Encode datetime for msgpack."""
    if isinstance(obj, datetime):
        obj = {"__datetime__": True, "as_str": obj.strftime("%Y%m%dT%H:%M:%S.%f")}
    return obj


class Storage(abc.ABC):
    """Storage class implementation.

    Args:
        serialized (typing.Optional[bytes], optional): Serialized data.
            Defaults to None.
        no_defaults (bool, optional): Storage meta data should have no defaults at
            initialization. Defaults to False.
    """

    def __init__(
        self, serialized: typing.Optional[bytes] = None, no_defaults: bool = False
    ) -> None:
        """Initialize Storage class."""
        if serialized:
            self.deserialize(serialized)
        else:
            self.data: np.ndarray = np.zeros((1024,), dtype=np.complex128)
            self.meta: Meta = Meta(no_defaults=no_defaults)

    @property
    def samples(self) -> int:
        """Number of waveform samples."""
        return len(self.data)

    def serialize(self) -> bytes:
        """Serialize content from storage."""
        if msgpack_import:
            content = pickle.dumps(self.__dict__, default=encode_datetime)
        else:
            content = pickle.dumps(self.__dict__)
        return content

    def deserialize(self, bins: bytes) -> None:
        """Deserialize content from binary."""
        if msgpack_import:
            content = pickle.loads(bins, object_hook=decode_datetime)
        else:
            content = pickle.loads(bins)
        self.__dict__ = content



"""Load implementation for iqw."""

from __future__ import annotations

import typing

import numpy as np

from ..load_interface import LoadInterface
from ..parent_storage import ParentStorage
from ..utility.fake_jit import jit
from ..utility.file_handling import read_file_handle

if typing.TYPE_CHECKING:
    from pathlib import Path


class Load(LoadInterface):
    """Load class for iqw."""

    def load(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load iwq data from file."""
        parent_storage = ParentStorage(no_defaults=True)
        with read_file_handle(file) as fp:
            content = fp.read()
        iq_data = self._extract_iq(content)
        parent_storage.storages[0].data = iq_data
        if isinstance(file, str):
            parent_storage.filename = file
        return parent_storage

    def load_in_chunks(
        self, file: typing.Union[str, typing.IO, Path], samples: int, offset: int
    ) -> ParentStorage:
        """Load chunk iwq data from file."""
        parent_storage = ParentStorage()
        with read_file_handle(file) as fp:
            fp.read(offset * 4 * 2)
            content = fp.read(samples * 4 * 2)
        iq_data = self._extract_iq(content)
        parent_storage.storages[0].data = iq_data
        if isinstance(file, str):
            parent_storage.filename = file
        return parent_storage

    def load_meta(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load meta information only from iqw file."""
        raise Exception("Iqw does not contain meta data!")

    @staticmethod
    def _extract_data(content: bytes):
        dt = np.dtype(np.float32)
        i_values = np.frombuffer(
            buffer=Load._extract_pairs(content, 0),
            dtype=dt,
        )
        q_values = np.frombuffer(
            buffer=Load._extract_pairs(content, 1),
            dtype=dt,
        )
        data = i_values + 1j * q_values
        return data

    def _extract_iq(self, content: bytes) -> np.ndarray:
        data = self._extract_data(content)
        return data

    @staticmethod
    @jit(forceobj=True)
    def _extract_pairs(buf: bytes, start: int = 0) -> bytes:
        """Extract data pairs from binary buffer."""
        single_pair = bytearray()
        length = len(buf)
        for idx in np.arange(start * 4, length, 8):
            single_pair.extend(buf[idx : idx + 4])
        return single_pair



"""Load implementation for iqtar."""

from __future__ import annotations

import os
import typing
import xml.etree.ElementTree as ET
from datetime import datetime
from io import BytesIO

import numpy as np

from RsWaveform.meta import Meta

from ..iqw import Load as LoadIqw
from ..load_interface import LoadInterface
from ..parent_storage import ParentStorage
from ..storage import Storage
from ..utility.file_handling import read_file_handle, read_file_handle_tar
from ..utility.meta_utils import map_meta_information_name

if typing.TYPE_CHECKING:
    from pathlib import Path


class Load(LoadInterface):
    """Load class for iqw."""

    def load(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load iq tar data from file."""
        parent_storage = self._read(file)
        if isinstance(file, str):
            parent_storage.filename = file
        return parent_storage

    def load_in_chunks(
        self, file: typing.Union[str, typing.IO, Path], samples: int, offset: int
    ) -> ParentStorage:
        """Load chunk iq tar data from file."""
        parent_storage = self._read_in_chunks(file, samples, offset)
        if isinstance(file, str):
            parent_storage.filename = file
        return parent_storage

    def load_meta(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load meta information only from iqtar file."""
        xml_content = self._extract_data_from_xml_file(file)
        tags = self._split_data_via_tags(xml_content)
        meta = self._extract_meta(tags)
        parent_storage = ParentStorage()
        parent_storage.storages[0].data = np.array([])
        parent_storage.storages[0].meta = Meta(**meta)
        return parent_storage

    @staticmethod
    def _allowed_tags() -> list:
        allowed_tags = [
            [["Clock"], "unit"],
            [["DataFilename"], None],
            [["Samples"], None],
            [["ScalingFactor"], "unit"],
            [["NumberOfChannels"], "unit"],
            [["DataType"], None],
            [["Format"], None],
            [["Name"], None],
            [["Comment"], None],
            [["DateTime"], None],
            [["UserData", "RohdeSchwarz", "SpectrumAnalyzer", "CenterFrequency"], None],
        ]
        return allowed_tags

    def _split_data_via_tags(self, content: bytes) -> dict:
        with BytesIO(content) as content_handle:
            tree = ET.parse(content_handle)
            root = tree.getroot()
            extracted: typing.Dict[str, typing.Any] = {}
            allowed_tags = self._allowed_tags()
            for tags, unit in allowed_tags:
                tag = "/".join(tags)
                search_tag = tag.replace("_", " ")
                meta_key = tags[-1].lower()
                xml_tag = root.find(search_tag)
                if xml_tag is not None:
                    value = xml_tag.text
                    if unit:
                        unit = xml_tag.get(unit)
                    else:
                        unit = None
                    extracted.update(
                        {map_meta_information_name(meta_key): [value, unit]}
                    )
            del root
        return extracted

    @staticmethod
    def _extract_meta(tags: dict) -> dict:
        meta = {}
        for key, [value, unit] in tags.items():
            if isinstance(value, bytes):
                value = value.decode()
            if key == "clock":
                value = float(value)
            elif key == "samples":
                continue
            elif key == "numberofchannels":
                value = int(value)
            elif key == "scalingfactor":
                value = float(value)
                if unit and unit != "V":
                    raise ValueError("unsupported unit!")
            elif key == "datafilename":
                value = str(value)
            elif key == "date":
                date_pattern = "%Y-%m-%dT%H:%M:%S" + (".%f" if "." in value else "")
                value = datetime.strptime(value, date_pattern)
            meta.update({key: value})
        return meta

    @staticmethod
    def _extract_data_from_xml_file(file: typing.Union[str, typing.IO, Path]) -> bytes:
        with read_file_handle_tar(file) as tar:
            files = tar.getnames()
            xml_filenames = [file for file in files if ".xml" in file.lower()]
            xml_filename = xml_filenames[0]
            tar.extract(xml_filename)
        with read_file_handle(xml_filename) as fp:
            content = fp.read()
        os.remove(xml_filename)
        return content

    @staticmethod
    def _extract_data_from_binary_file(
        file: typing.Union[str, typing.IO, Path], binary_filename: str
    ) -> bytes:
        with read_file_handle_tar(file) as tar:
            tar.extract(binary_filename)
        with read_file_handle(binary_filename) as fp:
            content = fp.read()
        os.remove(binary_filename)
        return content

    @staticmethod
    def _extract_data_from_binary_file_chunks(
        file: typing.Union[str, typing.IO, Path],
        binary_filename: str,
        nr_samples: int,
        samples_offset: int,
    ) -> bytes:
        with read_file_handle_tar(file) as tar:
            tar.extract(binary_filename)
        with read_file_handle(binary_filename) as fp:
            fp.read(samples_offset * 4 * 2)
            content = fp.read(nr_samples * 4 * 2)
        os.remove(binary_filename)
        return content

    @staticmethod
    def _extract_data(content: bytes) -> np.ndarray:
        parent_storage = ParentStorage()
        parent_storage.storages = []
        load_iqw = LoadIqw()
        with BytesIO(content) as file_handle:
            iq_storage = load_iqw.load(file_handle)
        return iq_storage.storages[0].data

    @staticmethod
    def _scale_data(data: np.ndarray, meta: dict) -> np.ndarray:
        scaling_factor = meta.get("scalingfactor", 1)
        if scaling_factor != 1:
            data = scaling_factor * data
        return data

    def _read(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        xml_content = self._extract_data_from_xml_file(file)

        tags = self._split_data_via_tags(xml_content)
        meta = self._extract_meta(tags)

        binary_file_name = meta.pop("datafilename")
        binary_content = self._extract_data_from_binary_file(file, binary_file_name)

        samples = self._extract_data(binary_content)
        samples = self._scale_data(samples, meta)

        parent_storage = ParentStorage()
        parent_storage.storages = []
        number_of_channels = meta.pop("numberofchannels", 1)
        for channel in range(number_of_channels):
            storage = Storage()
            samples_per_channel = int(len(samples) / number_of_channels)
            storage.data = samples[
                channel * samples_per_channel : samples_per_channel * (channel + 1)
            ]
            storage.meta = Meta(**meta)
            parent_storage.storages.append(storage)
        return parent_storage

    def _read_in_chunks(
        self,
        file: typing.Union[str, typing.IO, Path],
        nr_samples: int,
        samples_offset: int,
    ) -> ParentStorage:
        xml_content = self._extract_data_from_xml_file(file)

        tags = self._split_data_via_tags(xml_content)
        meta = self._extract_meta(tags)

        number_of_channels = meta.pop("numberofchannels", 1)
        if number_of_channels > 1:
            raise ValueError("Reading in chunks only supported for single channel!")

        binary_file_name = meta.pop("datafilename")
        binary_content = self._extract_data_from_binary_file_chunks(
            file, binary_file_name, nr_samples, samples_offset
        )

        samples = self._extract_data(binary_content)
        samples = self._scale_data(samples, meta)

        parent_storage = ParentStorage()
        parent_storage.storages = []

        storage = Storage()
        storage.data = samples
        storage.meta = Meta(**meta)
        parent_storage.storages.append(storage)
        return parent_storage



import pytest

from RsWaveform.utility.fake_jit import jit


def test_jit_with_func():
    try:
        import numba  # noqa F401
    except ImportError:
        pass
    else:
        pytest.skip("Numba is install but test should check decorator in no-numba mode")

    def dummy(*args):
        pass

    fn = jit(dummy)
    assert dummy is fn


def test_jit_with_signature():
    fn = jit("Int32,Int32")
    assert callable(fn)



from RsWaveform import __version__
from RsWaveform.__main__ import main


def test_main_prints_version(capsys):
    try:
        main(["--version"])
    except SystemExit:
        pass
    assert capsys.readouterr().out == f"{__version__}\n"



"""File handling utilities."""

from __future__ import annotations

import io
import tarfile
import typing
from contextlib import contextmanager

if typing.TYPE_CHECKING:
    from pathlib import Path


@contextmanager
def read_file_handle(
    file: typing.Union[str, typing.IO, Path]
) -> typing.Generator[typing.IO, None, None]:
    """Read from a waveform file."""
    return _file_handle(file, "rb")


@contextmanager
def write_file_handle(
    file: typing.Union[str, typing.IO, Path]
) -> typing.Generator[typing.IO, None, None]:
    """Write to a waveform file."""
    return _file_handle(file, "wb+")


def _file_handle(
    file: typing.Union[str, typing.IO, Path],
    mode: str,
    encoding: typing.Optional[str] = None,
) -> typing.Generator[typing.IO, None, None]:
    """File handler context manager implementation."""
    if isinstance(file, io.IOBase):
        yield file
    else:
        with open(typing.cast(str, file), mode, encoding=encoding) as f:
            yield f


@contextmanager
def read_file_handle_tar(
    file: typing.Union[str, tarfile.TarFile, typing.IO, Path]
) -> typing.Generator[tarfile.TarFile, None, None]:
    """Read from a tar file."""
    return _file_handle_tar(file, "r:")


@contextmanager
def write_file_handle_tar(
    file: typing.Union[str, tarfile.TarFile, typing.IO, Path]
) -> typing.Generator[tarfile.TarFile, None, None]:
    """Write to a tar file."""
    return _file_handle_tar(file, "w:")


def _file_handle_tar(
    file: typing.Union[str, tarfile.TarFile, typing.IO, Path], mode: str
) -> typing.Generator[tarfile.TarFile, None, None]:
    """File handler context manager implementation."""
    if isinstance(file, tarfile.TarFile):
        yield file
    elif isinstance(file, io.BytesIO):
        with tarfile.open(fileobj=file, mode=mode) as f:  # not perfect
            yield f
    else:
        with tarfile.open(typing.cast(str, file), mode) as f:
            yield f



"""Command line interface."""

import argparse

from . import __version__


def main(argv=None):
    """Parse arguments and execute commands."""
    create_parser().parse_args(argv)


def create_parser() -> argparse.ArgumentParser:
    """Create command line argument parser."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-V",
        "--version",
        help="Show version number and exit",
        action="version",
        version=__version__,
    )
    return parser


if __name__ == "__main__":
    main()



from datetime import datetime
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from RsWaveform.meta import Meta
from RsWaveform.npz import Load, Save
from RsWaveform.parent_storage import ParentStorage
from RsWaveform.storage import Storage


@pytest.fixture
def reference() -> np.ndarray:  # type: ignore
    return np.array([0.2, 0.6]) + 1j * np.array([0.4, 0.8])


@pytest.fixture
def meta() -> Meta:
    return Meta(
        **{  # type: ignore[arg-type]
            "type": "SMU-WV",
            "copyright": "Rohde & Schwarz",
            "comment": "Test waveform file",
            "control_length": 2,
            "date": datetime(2023, 1, 5, 10, 3, 52),
            "clock": 100000000.0,
            "rms": 2.220459,
            "peak": 0.0,
            "samples": 2,
            "reflevel": -20.0,
        }
    )


@patch("RsWaveform.npz.save.np")
def test_save(mock_np: MagicMock, reference, meta):
    mock_np.real = np.real
    mock_np.imag = np.imag
    mock_np.iinfo = np.iinfo
    parent_storage = ParentStorage()
    storage = Storage()
    storage.data = reference
    storage.meta = meta
    parent_storage.storages[0] = storage
    saver = Save()
    saver.dtype = np.float16
    filename = "test.npz"
    saver.save(filename, parent_storage)
    call_args = mock_np.savez_compressed.call_args[1]
    assert filename == call_args.get("file")
    assert np.array_equal(
        np.real(reference).astype(np.float16), call_args.get("storages")[0]["i"]
    )
    assert np.array_equal(
        np.imag(reference).astype(np.float16), call_args.get("storages")[0]["q"]
    )
    assert meta._items == call_args.get("storages")[0]["meta"]  # pylint: disable=W0212


def test_load(reference, meta, reference_npz_file_name):
    loader = Load()
    loader.dtype = np.float16
    parent_storage = loader.load(reference_npz_file_name)
    storage = parent_storage.storages[0]
    assert np.array_equal(
        np.real(storage.data).astype(np.float16), np.real(reference).astype(np.float16)
    )
    assert np.array_equal(
        np.imag(storage.data).astype(np.float16), np.imag(reference).astype(np.float16)
    )
    assert storage.meta == meta


@patch("RsWaveform.npz.load.np")
def test_load_not_schema(mock_np, reference_npz_file_name):
    mock_np.load.return_value.get.return_value = None
    loader = Load()
    loader.dtype = np.float16
    with pytest.raises(ValueError):
        loader.load(reference_npz_file_name)


@patch("RsWaveform.npz.load.np")
def test_load_skip_i_is_none(mock_np, reference_npz_file_name, caplog):
    mock_np.zeros = np.zeros
    mock_np.load.return_value.get.return_value = [{"i": None}]
    loader = Load()
    loader.dtype = np.float16
    loader.load(reference_npz_file_name)
    assert caplog.record_tuples[0] == (
        "RsWaveform.npz.load",
        30,
        (
            f"Skipping index 0 of file {reference_npz_file_name}"
            " content because no i values are in it."
        ),
    )


@patch("RsWaveform.npz.load.np")
def test_load_q_is_none(mock_np, reference_npz_file_name):
    mock_np.zeros = np.zeros
    real_values = np.array([1, 2])
    mock_np.load.return_value.get.return_value = [{"i": real_values, "q": None}]
    loader = Load()
    loader.dtype = np.float16
    parent_storage = loader.load(reference_npz_file_name)
    storage = parent_storage.storages[0]
    assert np.array_equal(real_values, storage.data)


@patch("RsWaveform.npz.load.np")
def test_load_skip_i_and_q_not_equal(mock_np, reference_npz_file_name, caplog):
    mock_np.zeros = np.zeros
    mock_np.load.return_value.get.return_value = [
        {"i": np.array([1, 2]), "q": np.array([3, 4, 5])}
    ]
    loader = Load()
    loader.dtype = np.float16
    loader.load(reference_npz_file_name)
    assert caplog.record_tuples[0] == (
        "RsWaveform.npz.load",
        30,
        (
            f"Skipping index 0 of file {reference_npz_file_name}"
            " content because i value and q value lengths are not identical."
        ),
    )


def test_load_in_chunks(reference, meta, reference_npz_file_name):
    loader = Load()
    loader.dtype = np.float16
    parent_storage = loader.load_in_chunks(reference_npz_file_name, 1, 0)
    storage = parent_storage.storages[0]
    assert np.array_equal(
        np.real(storage.data[:1]).astype(np.float16),
        np.real(reference[:1]).astype(np.float16),
    )
    assert np.array_equal(
        np.imag(storage.data[:1]).astype(np.float16),
        np.imag(reference[:1]).astype(np.float16),
    )
    assert storage.meta == meta
    with pytest.raises(ValueError):
        loader.load_in_chunks(reference_npz_file_name, 10, 0)


def test_load_meta(meta, reference_npz_file_name):
    loader = Load()
    loader.dtype = np.float16
    parent_storage = loader.load_meta(reference_npz_file_name)
    storage = parent_storage.storages[0]
    assert storage.meta == meta



"""Npz interface."""

import numpy as np
import numpy.typing as npt


class NpzInterface:
    """Interface class for Npz."""

    def __init__(self) -> None:
        """Instantiate interface class."""
        self._dtype: npt.DTypeLike = np.dtype(np.float16)

    @property
    def dtype(self) -> npt.DTypeLike:
        """Dtype for loading and saving.

        Returns:
            DTypeLike: Numpy dtype
        """
        return self._dtype

    @dtype.setter
    def dtype(self, value: npt.DTypeLike) -> None:
        """Dtype for loading and saving.

        Args:
            value (DTypeLike): Numpy dtype
        """
        self._dtype = value



"""Meta information utilities."""


def map_meta_information_name(key: str) -> str:
    """Map meta key."""
    mapping = {"datetime": "date"}
    if key in mapping:
        return mapping[key]
    else:
        return key



"""IQTAR meta data for Storage container."""

import typing

from .defaults import META_IQTAR_DEFAULTS
from .meta_base import MetaBase


class MetaIqtar(MetaBase):
    """IQTAR meta data for Storage container."""

    def __init__(self, no_defaults: bool = False, **kwargs):
        """Initialize IQTAR Meta class."""
        if no_defaults:
            values = {}
        else:
            values = META_IQTAR_DEFAULTS.copy()
        values.update(kwargs)
        MetaBase.__init__(self, values)

    @property
    def center_frequency(self) -> typing.Optional[str]:
        """Center frequency for IQTAR format."""
        return self._items.get("center_frequency")

    @center_frequency.setter
    def center_frequency(self, value: str) -> None:
        self._items["center_frequency"] = value

    @property
    def scalingfactor(self) -> typing.Optional[float]:
        """Scaling factor for IQTAR format."""
        return self._items.get("scalingfactor")

    @scalingfactor.setter
    def scalingfactor(self, value: float) -> None:
        self._items["scalingfactor"] = value



"""WV meta data for Storage container."""

import typing

import numpy as np

from .defaults import META_WV_DEFAULTS
from .meta_base import MetaBase


class MetaWv(MetaBase):
    """WV meta data for Storage container."""

    def __init__(self, no_defaults: bool = False, **kwargs):
        """Initialize WV Meta class."""
        if no_defaults:
            values = {}
        else:
            values = META_WV_DEFAULTS.copy()
        values.update(kwargs)
        MetaBase.__init__(self, values)

    @property
    def type(self) -> typing.Optional[str]:
        """Waveform type for WV format."""
        return self._items.get("type")

    @type.setter
    def type(self, value: str) -> None:
        self._items["type"] = value

    @property
    def copyright(self) -> typing.Optional[str]:
        """Coypright for WV format."""
        return self._items.get("copyright")

    @copyright.setter
    def copyright(self, value: str) -> None:
        self._items["copyright"] = value

    @property
    def marker(self) -> typing.Dict[str, typing.List[typing.List[int]]]:
        """Markers for WV format."""
        if "marker" not in self._items:
            self._items["marker"] = {}
        return self._items["marker"]

    @marker.setter
    def marker(self, value: typing.Dict[str, typing.List[typing.List[int]]]) -> None:
        self._items["marker"] = value

    @property
    def control_list(
        self,
    ) -> typing.Union[None, np.ndarray, typing.List[typing.List[int]]]:
        """Binary marker element stream for WV format.

        The list/np.ndarray format should be 'Marker 1-4 x Samples'

        Example:
        control_list = [[0, 1], [1, 0], [1, 0], [0, 1]]

        whereat Marker 1 looks like [0, 1], Marker 2 [1, 0], Marker 3 [1, 0],
        Marker 4 [0, 1].
        """
        return self._items.get("control_list")

    @control_list.setter
    def control_list(
        self, value: typing.Union[np.ndarray, typing.List[typing.List[int]]]
    ) -> None:
        self._items["control_list"] = value

    @property
    def control_length(self) -> typing.Optional[int]:
        """Control length for WV format."""
        return self._items.get("control_length")

    @control_length.setter
    def control_length(self, value: typing.Optional[int]) -> None:
        self._items["control_length"] = value

    @property
    def rms(self) -> typing.Optional[float]:
        """The RMS value of the signal."""
        return self._items.get("rms")

    @rms.setter
    def rms(self, value: float) -> None:
        self._items["rms"] = value

    @property
    def peak(self) -> typing.Optional[float]:
        """The signal peak value of the signal."""
        return self._items.get("peak")

    @peak.setter
    def peak(self, value: float) -> None:
        self._items["peak"] = value

    @property
    def samples(self) -> typing.Optional[int]:
        """The sample count of the waveform."""
        return self._items.get("samples")

    @samples.setter
    def samples(self, value: int) -> None:
        self._items["samples"] = value

    @property
    def reflevel(self) -> typing.Optional[float]:
        """The reference value of the signal."""
        return self._items.get("reflevel")

    @reflevel.setter
    def reflevel(self, value: float) -> None:
        self._items["reflevel"] = value



"""Meta data for Storage container."""

from .meta_iqtar import MetaIqtar
from .meta_wv import MetaWv


class Meta(MetaWv, MetaIqtar):
    """Meta data for Storage container.

    Args:
        no_defaults (bool, optional): Storage meta data should have no defaults at
            initialization. Defaults to False.
    """

    def __init__(self, no_defaults: bool = False, **kwargs):
        """Initialize Meta class."""
        MetaWv.__init__(self, no_defaults, **kwargs)
        MetaIqtar.__init__(self, no_defaults, **kwargs)



"""Save implementation for generic waveform."""

from __future__ import annotations

import typing

import numpy as np

from ..parent_storage import ParentStorage
from ..save_interface import SaveInterface
from .npz_inteface import NpzInterface

if typing.TYPE_CHECKING:
    from pathlib import Path


class Save(SaveInterface, NpzInterface):
    """Save class for generic waveform."""

    def save(
        self,
        file: typing.Union[str, typing.IO, Path],
        datas: ParentStorage,
        scale: float = np.power(2, np.iinfo(np.int16).bits - 1),
    ) -> None:
        """Save waveform data to file."""
        data_objs = []
        for storage in datas.storages:
            i_values = np.real(storage.data).astype(self.dtype)
            q_values = np.imag(storage.data).astype(self.dtype)
            meta = storage.meta._items  # pylint: disable=W0212
            data_objs.append({"i": i_values, "q": q_values, "meta": meta})
        np.savez_compressed(file=file, storages=data_objs)  # type: ignore



"""Convert arrays to a binary representation."""

import logging

import numpy as np

LOGGER = logging.getLogger(__name__)


def pack_bool_array_to_bytes(bool_array: np.ndarray) -> bytes:
    """Pack array values into a binary representation.

    Args:
        bool_array (np.ndarray): Array containing boolean values in integer
            representation.

    Raises:
        ValueError: If any of the number of rows or columns does not match 4.

    Returns:
        bytes: Packed list as bytes.
    """
    if len(bool_array.shape) != 2:
        raise ValueError("Array should have just two dimensions.")

    # Check if the number of rows is always 4.
    is_four = [x == 4 for x in bool_array.shape]
    if not any(is_four):
        raise ValueError("Number of rows must be four for packing into bytes")

    # Check if array is in column-wise major order, transpose if necessary
    if is_four[0]:
        bool_array = bool_array.T

    # We need an even number of samples for the columns
    if bool_array.shape[0] % 2 != 0:
        bool_array = np.concatenate([bool_array, np.zeros((1, 4))], axis=0)

    # Convert boolean array to integers (0 or 1)
    int_array = bool_array.astype(np.uint8)
    int_array = int_array.reshape((int(int_array.shape[0] / 2), 8))

    # Pack two columns into a single byte
    packed = np.packbits(int_array, axis=1, bitorder="big")
    packed_bytes = packed.tobytes()

    return packed_bytes


def unpack_bytes_to_bool_array(packed_bytes: bytes, num_samples: int) -> np.ndarray:
    """Unpack array values from a binary representation.

    Args:
        packed_bytes (bytes): Packed list as bytes.
        num_samples (int): The numnber of samples to extract.

    Returns:
        np.ndarray: The unpacked list.
    """
    # Unpack bytes to numpy array of integers (0 or 1)
    int_array = np.frombuffer(packed_bytes, dtype=np.uint8)

    # Unpack each byte to 8 bits (columns)
    bool_array = np.unpackbits(int_array, axis=0, bitorder="big")
    bool_array = bool_array.reshape(int(bool_array.shape[0] / 4), 4).T

    # Trim extra bits if the number of samples is not a multiple of 4
    bool_array = bool_array[:, :num_samples]

    return bool_array



"""A IndexedProperty that check for integer keys."""

from indexedproperty import IndexedProperty


class IntegerIndexedProperty(IndexedProperty):
    """A IndexedProperty that check for integer keys."""

    class _Trampoline(IndexedProperty._Trampoline):
        def moduserindex(self, index=0):
            """Modify user index."""
            if not isinstance(index, int):
                raise ValueError("Just integer as keys allowed.")
            return super().moduserindex(index)

        def modindex(self, index=0):
            """Modify index."""
            if not isinstance(index, int):
                raise ValueError("Just integer as keys allowed.")
            return super().modindex(index)



"""Defaults for WV meta data."""

META_WV_DEFAULTS = {
    "type": ("SMU-WV", 0),
    "copyright": "Rohde & Schwarz",
    "comment": "Created with RsWaveform",
    "clock": 1e9,
    "marker": {},
    "control_length": None,
    "control_list": None,
    "rms": None,
    "peak": None,
    "samples": None,
    "reflevel": None,
}

META_IQTAR_DEFAULTS = {
    "center_frequency": 1e9,
    "comment": "Created with RsWaveform",
    "clock": 1e9,
    "scalingfactor": 1,
}



"""Save implementation for iqtar."""

import os
import re
import typing
from pathlib import Path

from ..iqw.save import Save as SaveIqw
from ..meta.defaults import META_IQTAR_DEFAULTS as META_DEFAULTS
from ..parent_storage import ParentStorage
from ..save_interface import SaveInterface
from ..storage import Storage
from ..utility.file_handling import write_file_handle, write_file_handle_tar


class Save(SaveInterface):
    """Save class for iqtar."""

    def save(
        self,
        file: typing.Union[str, typing.IO, Path],
        datas: ParentStorage,
        scale: float = 1.0,
    ) -> None:
        """Save iq samples to iqtar file."""
        self._write(file, datas, scale)

    @staticmethod
    def _write_date(file: typing.IO, data: ParentStorage):
        date = data.timestamp
        line = f"<DateTime>{date.isoformat()}</DateTime>\n"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_comment(file: typing.IO, data: Storage):
        line = (
            f"<Comment>{data.meta.get('comment', META_DEFAULTS['comment'])}"
            f"</Comment>\n"
        )
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_samples(file: typing.IO, data: Storage):
        line = f"<Samples>{len(data.data)}</Samples>\n"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_clock(file: typing.IO, data: Storage):
        clock = data.meta.get("clock")
        if not clock:
            raise ValueError("Clock is a mandatory parameter!")
        line = f'<Clock unit="Hz">{clock}</Clock>'
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_center_frequency(file: typing.IO, data: Storage):
        if data.meta.get("center_frequency", 0) != 0:
            file.write(
                (
                    f"<UserData><RohdeSchwarz>"
                    f"<SpectrumAnalyzer>"
                    f'<CenterFrequency unit="Hz">'
                    f'{data.meta.get("center_frequency")}'
                    f"</CenterFrequency>"
                    f"</SpectrumAnalyzer>"
                    f"</RohdeSchwarz></UserData>\n"
                ).encode("utf-8")
            )

    @staticmethod
    def _write_number_of_channels(file: typing.IO, data: ParentStorage):
        line = f"<NumberOfChannels>{data.number_of_storages()}</NumberOfChannels>\n"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_overhead(file: typing.IO):
        file.write('<?xml version="1.0" encoding="UTF-8"?>\n'.encode("utf-8"))
        file.write(
            '<?xml-stylesheet type="text/xsl" '
            'href="open_IqTar_xml_file_in_web_browser.xslt"?>\n'.encode("utf-8")
        )
        file.write(
            "<RS_IQ_TAR_FileFormat "
            'fileFormatVersion="2" '
            "xsi:noNamespaceSchemaLocation="
            '"http://www.rohde-schwarz.com/file/RsIqTar.xsd" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            "\n".encode("utf-8")
        )
        file.write("<Name>Python iq.tar Writer</Name>\n".encode("utf-8"))

    @staticmethod
    def _write_datafilename(file: typing.IO, filenameiqw):
        line = "<DataFilename>" + filenameiqw + "</DataFilename>\n"
        file.write(line.encode("utf-8"))

    @staticmethod
    def _write_scaling_factor(file: typing.IO, data: Storage):
        line = (
            f'<ScalingFactor unit="V">'
            f"{data.meta.get('scalingfactor', 1)}</ScalingFactor>\n"
        )
        file.write(line.encode("utf-8"))
        pass

    def _write_xml(
        self,
        data: ParentStorage,
        filename_xml: typing.Union[str, Path],
        filename_iqw: str,
    ):
        with write_file_handle(filename_xml) as xml_file:
            self._write_overhead(xml_file)
            self._write_date(xml_file, data)
            self._write_comment(xml_file, data.storages[0])
            self._write_samples(xml_file, data.storages[0])
            self._write_clock(xml_file, data.storages[0])
            xml_file.write("<Format>complex</Format>\n".encode("utf-8"))
            xml_file.write("<DataType>float32</DataType>\n".encode("utf-8"))
            self._write_scaling_factor(xml_file, data.storages[0])
            self._write_datafilename(xml_file, filename_iqw)
            self._write_number_of_channels(xml_file, data)
            self._write_center_frequency(xml_file, data.storages[0])
            xml_file.write("</RS_IQ_TAR_FileFormat>\n".encode("utf-8"))

    @staticmethod
    def _write_data(
        binaryfile: typing.Union[str, Path], data: ParentStorage, scale: float = 1.0
    ):
        save_iqw = SaveIqw()
        save_iqw.save(binaryfile, data, scale)

    def _write(
        self,
        file: typing.Union[str, typing.IO, Path],
        data: ParentStorage,
        scale: float = 1.0,
    ):
        if isinstance(file, str):
            filename = Path(file).name
            path = Path(file).parent
            binaryfile = re.sub(
                "iq.tar", "complex.1ch.float32", filename, flags=re.IGNORECASE
            )
            xml_filename = re.sub("iq.tar", "xml", filename, flags=re.IGNORECASE)
        else:
            path = Path.home() / ".RsWaveform"
            binaryfile = "data.complex.1ch.float32"
            xml_filename = "data.xml"

        self._write_data(path / binaryfile, data, scale)
        self._write_xml(data, path / xml_filename, binaryfile)

        with write_file_handle_tar(file) as tar:
            tar.add(str(path / binaryfile), arcname=binaryfile)
            tar.add(str(path / xml_filename), arcname=xml_filename)
        os.remove(os.path.join(path, binaryfile))
        os.remove(os.path.join(path, xml_filename))



"""Save implementation for iqw."""

from __future__ import annotations

import math
import typing

import numpy as np

from ..parent_storage import ParentStorage
from ..save_interface import SaveInterface
from ..storage import Storage
from ..utility.fake_jit import jit
from ..utility.file_handling import write_file_handle

if typing.TYPE_CHECKING:
    from pathlib import Path


class Save(SaveInterface):
    """Save class for iqw."""

    def save(
        self,
        file: typing.Union[str, typing.IO, Path],
        datas: ParentStorage,
        scale: float = 1.0,
    ) -> None:
        """Save iq samples to iqw file."""
        self._write(file, datas, scale)

    @staticmethod
    @jit(forceobj=True)
    def _scale_data(iq_samples: np.ndarray, scale: float = 1.0) -> np.ndarray:
        iq_samples = (1 / scale) * iq_samples
        return iq_samples

    @staticmethod
    def _prepare_data(iq_samples: np.ndarray) -> np.ndarray:
        def get_real_or_complex_by_index(iq: complex, index: int):
            return iq.real if index == 0 else iq.imag

        sorted_iq_samples = [
            get_real_or_complex_by_index(iq, i) for iq in iq_samples for i in range(2)
        ]
        return np.asarray(sorted_iq_samples)

    def _write(
        self,
        file: typing.Union[str, typing.IO, Path],
        data: ParentStorage,
        scale: float = 1.0,
    ):
        with write_file_handle(file) as fp:
            for storage in data.storages:
                self._write_data(fp, storage, scale)

    def _write_data(self, file: typing.IO, data: Storage, scale: float = 1.0):
        scaled_iq_data = self._scale_data(data.data, scale)
        iq_data = self._prepare_data(scaled_iq_data)

        chunk_size = 82000000  # randomly chosen number. Might be changed!
        number_blocks = math.ceil(iq_data.size / chunk_size)
        for block_idx in range(number_blocks):
            idx_start = block_idx * chunk_size
            idx_stop = (block_idx + 1) * chunk_size
            if idx_stop > iq_data.size - 1:
                idx_stop = iq_data.size
            bins = iq_data[idx_start:idx_stop].astype(np.float32).tobytes()
            file.write(bins)



"""Meta base class implementation."""

import typing

META_DATA_DICT = ["_items", "meta_attribute"]


class MetaBase:
    """Common meta data for Storage container."""

    def __init__(self, defaults: typing.Optional[dict] = None) -> None:
        """Initialize Base Meta class."""
        if defaults is None:
            defaults = {}
        if not hasattr(self, "_items"):
            self._items = {}
        for key, value in defaults.items():
            self._items[key] = value

    def __getitem__(self, key):
        """Set item to internal dict."""
        return self._items[key]

    def __setitem__(self, key, value):
        """Get item from internal dict."""
        self._items[key] = value

    def __delitem__(self, key):
        """Delete item from internal dict."""
        del self._items[key]

    def __contains__(self, key):
        """Check if a key is in the dictionary."""
        return key in self._items

    def __iter__(self):
        """Iterate over the keys in the dictionary."""
        return iter(self._items)

    def __len__(self):
        """Get the number of key-value pairs."""
        return len(self._items)

    def __repr__(self):
        """Write the representation of Meta."""
        return self._items.__repr__()

    def __eq__(self, other):
        """Compare two MetaBase objects for equality."""
        if isinstance(other, MetaBase):
            return self._items == other._items
        if isinstance(other, dict):
            return self._items == other
        return False

    def __ne__(self, other):
        """Check if two MetaBase objects are not equal."""
        return not self == other

    def pop(self, key, default=None):
        """Remove a key and return its value. If key is not found, return default."""
        return self._items.pop(key, default)

    def popitem(self):
        """Remove and return a (key, value) pair as a 2-tuple.

        Pairs are returned in LIFO order.
        """
        return self._items.popitem()

    def clear(self):
        """Remove all items from the dictionary."""
        self._items.clear()

    def setdefault(self, key, default=None):
        """If key is in the dictionary, return its value.

        If not, insert key with a value of default and return default.
        """
        return self._items.setdefault(key, default)

    def copy(self):
        """Return a shallow copy of the dictionary."""
        instance_type = type(self)
        kwargs = self._items.copy()
        kwargs["no_defaults"] = True
        return instance_type(**kwargs)

    def keys(self):
        """Get the keys in the dictionary."""
        return self._items.keys()

    def values(self):
        """Get the values in the dictionary."""
        return self._items.values()

    def items(self):
        """Get the key-value pairs in the dictionary."""
        return self._items.items()

    def get(self, key, default=None):
        """Get the value for a key, or a default value if the key is not in the dict."""
        return self._items.get(key, default)

    def update(self, *args, **kwargs):
        """Update the dictionary with the key-value pairs from another dictionary."""
        return self._items.update(*args, **kwargs)

    # Shared properties ###

    @property
    def comment(self) -> typing.Optional[str]:
        """Mandatory parameter for all types of waveforms."""
        return self._items.get("comment")

    @comment.setter
    def comment(self, value: str) -> None:
        """Setter for comment."""
        self._items["comment"] = value

    @property
    def clock(self) -> typing.Optional[float]:
        """Mandatory parameter for all types of waveforms."""
        return self._items.get("clock")

    @clock.setter
    def clock(self, value: float) -> None:
        """Setter for clock."""
        self._items["clock"] = value



"""Load implementation for generic waveform."""

from __future__ import annotations

import logging
import typing

import numpy as np

from ..load_interface import LoadInterface
from ..meta import Meta
from ..parent_storage import ParentStorage
from ..storage import Storage
from .npz_inteface import NpzInterface

if typing.TYPE_CHECKING:
    from pathlib import Path


LOGGER = logging.getLogger(__name__)


class Load(LoadInterface, NpzInterface):
    """Load class for generic waveform."""

    def load(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load waveform data from npz file."""
        content = np.load(file=file, allow_pickle=True)
        data_objs = content.get("storages")
        if not data_objs:
            raise ValueError(f"File {file} is not compliant to data schema.")
        parent = ParentStorage(number_of_storages=len(data_objs))
        for idx, obj in enumerate(data_objs):
            i_values = obj.get("i")
            q_values = obj.get("q")
            meta = obj.get("meta", {})
            if i_values is None:
                LOGGER.warning(
                    (
                        "Skipping index %s of file %s content because no i values are"
                        " in it."
                    ),
                    idx,
                    str(file),
                )
                continue
            if q_values is None:
                q_values = np.zeros(i_values.size)
            if q_values is not None and len(i_values) != len(q_values):
                LOGGER.warning(
                    (
                        "Skipping index %s of file %s content because i value and"
                        " q value lengths are not identical."
                    ),
                    idx,
                    str(file),
                )
                continue
            storage = Storage()
            storage.data = i_values.astype(self.dtype) + 1j * q_values.astype(
                self.dtype
            )
            storage.meta = Meta(no_defaults=True, **meta)
            parent.storages[idx] = storage
        return parent

    def load_in_chunks(
        self,
        file: typing.Union[str, typing.IO, Path],
        samples: int,
        offset: int,
    ) -> ParentStorage:
        """Load chunk waveform data from file."""
        parent = self.load(file)
        for idx, storage in enumerate(parent.storages):
            if offset + samples > len(storage.data):
                raise ValueError(
                    f"Could not load chunks for index {idx} of file {file} because "
                    f"offset {offset} and sample count {samples} is greater than length"
                    f" of available data {len(storage.data)}. "
                )
            storage.data = storage.data[offset : offset + samples]
        return parent

    def load_meta(self, file: typing.Union[str, typing.IO, Path]) -> ParentStorage:
        """Load meta information only from wv file."""
        parent = self.load(file)
        for storage in parent.storages:
            storage.data = np.empty((0, 0))
        return parent



import pytest

from RsWaveform.utility.integer_indexedproperty import IntegerIndexedProperty


class DummyClass:
    def __init__(self, populate: int = 1):
        self._list = list(range(populate))

    @IntegerIndexedProperty
    def content(self, index: int = 0) -> int:
        return self._list[index]

    @content.setter
    def content_setter(self, index: int, value: int) -> None:
        self._list[index] = value


def test_moduserindex_valid_integer():
    prop = DummyClass()
    prop.content[0] = 42
    assert prop.content[0] == 42


def test_out_of_range_index():
    prop = DummyClass()

    with pytest.raises(IndexError):
        prop.content[1]


def test_invalid_non_integer():
    prop = DummyClass()

    with pytest.raises(ValueError, match="Just integer as keys allowed."):
        prop.content[1.5]

    with pytest.raises(ValueError, match="Just integer as keys allowed."):
        prop.content[1.5] = 5



"""Fake numba jit decorator."""

try:
    from numba import jit as jit
except ImportError:

    def jit(
        signature_or_function=None,
        locals={},
        cache=False,
        pipeline_class=None,
        boundscheck=None,
        **options,
    ):
        """Fake decorator for numba."""
        if signature_or_function is not None and callable(signature_or_function):
            return signature_or_function
        else:
            return lambda fn: fn
